import os
import time
import PySimpleGUI as sg
import pandas as pd
from pandas import DataFrame
import numpy as np
import matplotlib.pyplot as plt
from matplotlib_venn import venn2
from matplotlib_venn import venn3
from matplotlib.pyplot import plot, ion, show
from pathlib import Path
from taxontabletools.taxontable_manipulation import strip_metadata
from taxontabletools.create_log import ttt_log

file_a = '/Users/tillmacher/Desktop/TTT_projects/Projects/Lippe_eRNA_fwh_invertebrates/TaXon_tables/Lippe_eRNA_fwh_taxon_table_cons_NCsub_invertebrates_normalized_eDNA.xlsx'
file_b = '/Users/tillmacher/Desktop/TTT_projects/Projects/Lippe_eRNA_fwh_invertebrates/TaXon_tables/Lippe_eRNA_fwh_taxon_table_cons_NCsub_invertebrates_normalized_eRNA.xlsx'
file_c = ''
venn_diagram_name = 'test'
path_to_outdirs = 'test'
clustering_unit = 'OTUs'

def venn_diagram(file_a, file_b, file_c, venn_diagram_name, path_to_outdirs, clustering_unit):

    file_a = Path(file_a)
    file_b = Path(file_b)
    venn_font = 40

    if file_c == False:
    ############################################################################
    # use venn2

        count = 0

        G = "G_" + clustering_unit
        allowed_taxa = ["A_Phylum","B_Class","C_Order","D_Family","E_Genus","F_Species", G]
        venn_dict = {}
        data_file_a = pd.read_excel(file_a, header=0).fillna('')
        data_file_b = pd.read_excel(file_b, header=0).fillna('')

        for taxon in allowed_taxa:

            time.sleep(1)

            output_name = taxon
            taxon = taxon[2:]
            col_name = taxon

            if taxon in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
                col_name = taxon
                taxon="ID"

            file_name_a = file_a.stem
            file_name_b = file_b.stem

            taxa_file_a = set([i for i in data_file_a[taxon].values.tolist() if i != ''])
            taxa_file_b = set([i for i in data_file_b[taxon].values.tolist() if i != ''])

            a_only = taxa_file_a - taxa_file_b
            len_a_only = len(a_only)
            b_only = taxa_file_b - taxa_file_a
            len_b_only = len(b_only)
            shared = taxa_file_a & taxa_file_b
            len_shared = len(shared)

            venn_dict[col_name + "_a_only"] = a_only
            venn_dict[col_name + "_shared"] = shared
            venn_dict[col_name + "_b_only"] = b_only

            plt.figure(figsize=(20, 10))
            out = venn2(subsets = (len_a_only, len_b_only, len_shared), set_labels = (file_name_a, file_name_b))
            for text in out.set_labels:
                text.set_fontsize(venn_font)
            for x in range(len(out.subset_labels)):
                if out.subset_labels[x] is not None:
                    out.subset_labels[x].set_fontsize(venn_font)

            dirName = Path(str(path_to_outdirs) + "/Venn_diagrams/" + venn_diagram_name)
            if not os.path.exists(dirName):
                os.mkdir(dirName)

            output_pdf = Path(str(dirName) + "/" + output_name + ".pdf")
            plt.title(output_name[2:])
            plt.savefig(output_pdf, bbox_inches='tight')
            plt.close()

        output_xlsx = Path(str(dirName) + "/" + "Venn_comparison_results.xlsx")
        df = pd.DataFrame.from_dict(venn_dict, orient='index').transpose()
        df.to_excel(output_xlsx, index=False)

        sg.Popup("Venn diagrams are found in", path_to_outdirs, "Venn_diagrams/", title="Finished", keep_on_top=True)

        ttt_log("venn diagram", "analysis", file_a.name, output_xlsx.name, venn_diagram_name, path_to_outdirs)
        ttt_log("venn diagram", "analysis", file_b.name, output_xlsx.name, venn_diagram_name, path_to_outdirs)

    else:
        ############################################################################
        # use venn3

        if file_c == '':
            sg.PopupError("Please provide a file", keep_on_top=True)
            raise RuntimeError()

        file_c = Path(file_c)

        count = 0

        G = "G_" + clustering_unit
        allowed_taxa = ["A_Phylum","B_Class","C_Order","D_Family","E_Genus","F_Species", G]

        venn_dict = {}

        data_file_a = pd.read_excel(file_a, header=0).fillna('')
        data_file_b = pd.read_excel(file_b, header=0).fillna('')
        data_file_c = pd.read_excel(file_c, header=0).fillna('')

        for taxon in allowed_taxa:

            time.sleep(1)

            output_name = taxon
            taxon = taxon[2:]
            col_name = taxon

            if taxon in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
                col_name = taxon
                taxon="ID"

            file_name_a = file_a.stem
            file_name_b = file_b.stem
            file_name_c = file_c.stem

            taxa_file_a = data_file_a[taxon].values.tolist()
            taxa_file_b = data_file_b[taxon].values.tolist()
            taxa_file_c = data_file_c[taxon].values.tolist()

            taxa_unique_a = list(dict.fromkeys(taxa_file_a))
            taxa_unique_b = list(dict.fromkeys(taxa_file_b))
            taxa_unique_c = list(dict.fromkeys(taxa_file_c))

            taxa_labels_a = []
            taxa_labels_b = []
            taxa_labels_c = []
            taxa_sizes_a = []
            taxa_sizes_b = []
            taxa_sizes_c = []

            for taxon_name in taxa_unique_a:
                if "nan" != str(taxon_name):
                    taxa_labels_a.append(str(taxon_name))
                    taxa_sizes_a.append(taxa_file_a.count(taxon_name))

            for taxon_name in taxa_unique_b:
                if "nan" != str(taxon_name):
                    taxa_labels_b.append(str(taxon_name))
                    taxa_sizes_b.append(taxa_file_b.count(taxon_name))

            for taxon_name in taxa_unique_c:
                if "nan" != str(taxon_name):
                    taxa_labels_c.append(str(taxon_name))
                    taxa_sizes_c.append(taxa_file_c.count(taxon_name))

            taxa_labels_a = sorted(taxa_labels_a)
            taxa_labels_b = sorted(taxa_labels_b)
            taxa_labels_c = sorted(taxa_labels_c)

            a_only = set(taxa_labels_a) - set(taxa_labels_b) - set(taxa_labels_c)
            len_a_only = len(a_only)
            b_only = set(taxa_labels_b) - set(taxa_labels_a) - set(taxa_labels_c)
            len_b_only = len(b_only)
            c_only = set(taxa_labels_c) - set(taxa_labels_a) - set(taxa_labels_b)
            len_c_only = len(c_only)

            shared_all = set(taxa_labels_a) & set(taxa_labels_b) & set(taxa_labels_c)
            len_shared_all = len(shared_all)
            shared_a_b = set(taxa_labels_a) & set(taxa_labels_b) - set(taxa_labels_c)
            len_shared_a_b = len(shared_a_b)
            shared_a_c = set(taxa_labels_a) & set(taxa_labels_c) - set(taxa_labels_b)
            len_shared_a_c = len(shared_a_c)
            shared_b_c = set(taxa_labels_b) & set(taxa_labels_c) - set(taxa_labels_a)
            len_shared_b_c = len(shared_b_c)

            venn_dict[col_name + "_a_only"] = a_only
            venn_dict[col_name + "_b_only"] = b_only
            venn_dict[col_name + "_c_only"] = c_only
            venn_dict[col_name + "_shared_all"] = shared_all
            venn_dict[col_name + "_shared_a_b"] = shared_a_b
            venn_dict[col_name + "_shared_a_c"] = shared_a_c
            venn_dict[col_name + "_shared_b_c"] = shared_b_c

            plt.figure(figsize=(20, 10))
            out = venn3(subsets = (len_a_only, len_b_only, len_shared_a_b, len_c_only, len_shared_a_c, len_shared_b_c, len_shared_all), set_labels = (file_name_a, file_name_b, file_name_c))
            for text in out.set_labels:
                text.set_fontsize(venn_font)
            for x in range(len(out.subset_labels)):
                if out.subset_labels[x] is not None:
                    out.subset_labels[x].set_fontsize(venn_font)

            dirName = Path(str(path_to_outdirs) + "/Venn_diagrams/" + venn_diagram_name)
            if not os.path.exists(dirName):
                os.mkdir(dirName)

            output_pdf = Path(str(dirName) + "/" + output_name + ".pdf")
            plt.title(output_name[2:])
            plt.savefig(output_pdf, bbox_inches='tight')
            plt.close()

        output_xlsx = Path(str(dirName) + "/" + "Venn_comparison_results.xlsx")
        df = pd.DataFrame.from_dict(venn_dict, orient='index').transpose()
        df.to_excel(output_xlsx, index=False)

        sg.Popup("Venn diagrams are found in", path_to_outdirs, "Venn_diagrams/", title="Finished", keep_on_top=True)

        ttt_log("venn diagram", "analysis", file_a.name, output_xlsx.name, venn_diagram_name, path_to_outdirs)
        ttt_log("venn diagram", "analysis", file_b.name, output_xlsx.name, venn_diagram_name, path_to_outdirs)
        ttt_log("venn diagram", "analysis", file_c.name, output_xlsx.name, venn_diagram_name, path_to_outdirs)

def venn_diagram_metadata(TaXon_table_xlsx, venn_diagram_name, meta_data_to_test, path_to_outdirs, clustering_unit):

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('unidentified')
    TaXon_table_df = strip_metadata(TaXon_table_df)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0).fillna("nan")
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    metadata_loc = Meta_data_table_df.columns.tolist().index(meta_data_to_test)
    ## drop samples with metadata called nan (= empty)
    drop_samples = [i[0] for i in Meta_data_table_df.values.tolist() if i[metadata_loc] == "nan"]

    if drop_samples != []:
        ## filter the TaXon table
        TaXon_table_df = TaXon_table_df.drop(drop_samples, axis=1)
        TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
        ## also remove empty OTUs
        row_filter_list = []
        for row in TaXon_table_df.values.tolist():
            reads = set(row[10:])
            if reads != {0}:
                row_filter_list.append(row)
        columns = TaXon_table_df.columns.tolist()
        TaXon_table_df = pd.DataFrame(row_filter_list, columns=columns)
        Meta_data_table_df = pd.DataFrame([i for i in Meta_data_table_df.values.tolist() if i[0] not in drop_samples], columns=Meta_data_table_df.columns.tolist())
        Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    categories = Meta_data_table_df[meta_data_to_test].tolist()
    u_categories = list(set(categories))
    len_categories = len(u_categories)

    categories_dict = {}
    for sample, metadata in Meta_data_table_df[["Samples", meta_data_to_test]].values.tolist():
        if metadata in categories_dict.keys():
            categories_dict[metadata] = categories_dict[metadata] + [sample]
        else:
            categories_dict[metadata] = [sample]

    ## create a y axis title text
    taxon_title = clustering_unit.lower()

    ## adjust taxonomic level if neccessary
    if clustering_unit in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = clustering_unit
        clustering_unit = "ID"

    if len(set(Meta_data_table_df[meta_data_to_test])) == 1:
        sg.PopupError("The meta data has to differ between samples!", title="Error")
        raise RuntimeError

    if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):
        venn_font = 40

        if len_categories == 2:
        ############################################################################
        # use venn2

            count = 0

            G = "G_" + clustering_unit
            allowed_taxa = ["A_Phylum","B_Class","C_Order","D_Family","E_Genus","F_Species", G]

            venn_dict = {}

            for taxon in allowed_taxa:

                time.sleep(1)

                output_name = taxon
                taxon = taxon[2:]
                col_name = taxon

                if taxon in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
                    col_name = taxon
                    taxon="ID"

                ## search present taxa for each metadata
                samples_a = categories_dict[u_categories[0]]
                category_a = u_categories[0]
                taxa_labels_a = sorted(list(set([i[0] for i in TaXon_table_df[[taxon] + samples_a].values.tolist() if i[0] != 'unidentified' and sum(i[1::]) != 0])))

                samples_b = categories_dict[u_categories[1]]
                taxa_labels_b = sorted(list(set([i[0] for i in TaXon_table_df[[taxon] + samples_b].values.tolist() if i[0] != 'unidentified' and sum(i[1::]) != 0])))
                category_b = u_categories[1]

                a_only = set(taxa_labels_a) - set(taxa_labels_b)
                len_a_only = len(a_only)
                b_only = set(taxa_labels_b) - set(taxa_labels_a)
                len_b_only = len(b_only)
                shared = set(taxa_labels_a) & set(taxa_labels_b)
                len_shared = len(shared)

                venn_dict[col_name + "_a_only"] = a_only
                venn_dict[col_name + "_shared"] = shared
                venn_dict[col_name + "_b_only"] = b_only

                plt.figure(figsize=(20, 10))
                out = venn2(subsets = (len_a_only, len_b_only, len_shared), set_labels = (category_a, category_b))
                for text in out.set_labels:
                    text.set_fontsize(venn_font)
                for x in range(len(out.subset_labels)):
                    if out.subset_labels[x] is not None:
                        out.subset_labels[x].set_fontsize(venn_font)

                dirName = Path(str(path_to_outdirs) + "/Venn_diagrams/" + venn_diagram_name)
                if not os.path.exists(dirName):
                    os.mkdir(dirName)

                output_pdf = Path(str(dirName) + "/" + output_name + ".pdf")
                plt.title(output_name[2:])
                plt.savefig(output_pdf, bbox_inches='tight')

                plt.close()

            output_xlsx = Path(str(dirName) + "/" + "Venn_comparison_results.xlsx")
            df = pd.DataFrame.from_dict(venn_dict, orient='index').transpose()
            df.to_excel(output_xlsx, index=False)

            sg.Popup("Venn diagrams are found in", path_to_outdirs, "Venn_diagrams/", title="Finished", keep_on_top=True)

            ttt_log("venn diagram", "analysis", category_a, output_xlsx.name, venn_diagram_name, path_to_outdirs)
            ttt_log("venn diagram", "analysis", category_b, output_xlsx.name, venn_diagram_name, path_to_outdirs)

        elif len_categories == 3:
        ############################################################################
        # use venn3

            count = 0

            G = "G_" + clustering_unit
            allowed_taxa = ["A_Phylum","B_Class","C_Order","D_Family","E_Genus","F_Species", G]

            venn_dict = {}

            for taxon in allowed_taxa:

                time.sleep(1)

                output_name = taxon
                taxon = taxon[2:]
                col_name = taxon

                if taxon in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
                    col_name = taxon
                    taxon="ID"

                ## search present taxa for each metadata
                samples_a = categories_dict[u_categories[0]]
                category_a = u_categories[0]
                taxa_labels_a = sorted(list(set([i[0] for i in TaXon_table_df[[taxon] + samples_a].values.tolist() if i[0] != 'unidentified' and sum(i[1::]) != 0])))

                samples_b = categories_dict[u_categories[1]]
                category_b = u_categories[1]
                taxa_labels_b = sorted(list(set([i[0] for i in TaXon_table_df[[taxon] + samples_b].values.tolist() if i[0] != 'unidentified' and sum(i[1::]) != 0])))

                samples_c = categories_dict[u_categories[2]]
                category_c = u_categories[2]
                taxa_labels_c = sorted(list(set([i[0] for i in TaXon_table_df[[taxon] + samples_c].values.tolist() if i[0] != 'unidentified' and sum(i[1::]) != 0])))

                a_only = set(taxa_labels_a) - set(taxa_labels_b) - set(taxa_labels_c)
                len_a_only = len(a_only)
                b_only = set(taxa_labels_b) - set(taxa_labels_a) - set(taxa_labels_c)
                len_b_only = len(b_only)
                c_only = set(taxa_labels_c) - set(taxa_labels_a) - set(taxa_labels_b)
                len_c_only = len(c_only)

                shared_all = set(taxa_labels_a) & set(taxa_labels_b) & set(taxa_labels_c)
                len_shared_all = len(shared_all)
                shared_a_b = set(taxa_labels_a) & set(taxa_labels_b) - set(taxa_labels_c)
                len_shared_a_b = len(shared_a_b)
                shared_a_c = set(taxa_labels_a) & set(taxa_labels_c) - set(taxa_labels_b)
                len_shared_a_c = len(shared_a_c)
                shared_b_c = set(taxa_labels_b) & set(taxa_labels_c) - set(taxa_labels_a)
                len_shared_b_c = len(shared_b_c)

                venn_dict[col_name + "_a_only"] = a_only
                venn_dict[col_name + "_b_only"] = b_only
                venn_dict[col_name + "_c_only"] = c_only
                venn_dict[col_name + "_shared_all"] = shared_all
                venn_dict[col_name + "_shared_a_b"] = shared_a_b
                venn_dict[col_name + "_shared_a_c"] = shared_a_c
                venn_dict[col_name + "_shared_b_c"] = shared_b_c

                plt.figure(figsize=(20, 10))
                out = venn3(subsets = (len_a_only, len_b_only, len_shared_a_b, len_c_only, len_shared_a_c, len_shared_b_c, len_shared_all), set_labels = (category_a, category_b, category_c))
                for text in out.set_labels:
                    text.set_fontsize(venn_font)
                for x in range(len(out.subset_labels)):
                    if out.subset_labels[x] is not None:
                        out.subset_labels[x].set_fontsize(venn_font)

                dirName = Path(str(path_to_outdirs) + "/Venn_diagrams/" + venn_diagram_name)
                if not os.path.exists(dirName):
                    os.mkdir(dirName)

                output_pdf = Path(str(dirName) + "/" + output_name + ".pdf")
                plt.title(output_name[2:])
                plt.savefig(output_pdf, bbox_inches='tight')

                plt.close()

            output_xlsx = Path(str(dirName) + "/" + "Venn_comparison_results.xlsx")
            df = pd.DataFrame.from_dict(venn_dict, orient='index').transpose()
            df.to_excel(output_xlsx, index=False)

            sg.Popup("Venn diagrams are found in", str(path_to_outdirs), "Venn_diagrams/", title="Finished", keep_on_top=True)

            ttt_log("venn diagram", "analysis", category_a, output_xlsx.name, venn_diagram_name, path_to_outdirs)
            ttt_log("venn diagram", "analysis", category_b, output_xlsx.name, venn_diagram_name, path_to_outdirs)
            ttt_log("venn diagram", "analysis", category_c, output_xlsx.name, venn_diagram_name, path_to_outdirs)

        else:
            sg.PopupError("You cannot use more than 3 metadata categories for Venn diagrams!")

    else:
        sg.PopupError("Error: The samples do not match", title="Error")



import requests_html, json, datetime
import re, sys, subprocess, os
import PySimpleGUI as sg
import pandas as pd
from tqdm import tqdm
from statistics import mean
from pandas import DataFrame
from datetime import date
import numpy as np
from pathlib import Path
from Bio import pairwise2
from Bio.Seq import Seq
from taxontabletools.taxontable_manipulation import strip_metadata
from taxontabletools.taxontable_manipulation import collect_metadata
from taxontabletools.taxontable_manipulation import add_metadata

def create_taxon_list(TaXon_table_xlsx, taxon_list_output_file_name, create_gbif_link, calc_dist, use_metadata, taxon_tools_version, path_to_outdirs, clustering_unit):

    ##### functions
    def open_table(table):
        if sys.platform == "win32":
            os.startfile(table)
        else:
            opener = "open" if sys.platform == 'darwin' else 'xdg-open'
            subprocess.call([opener, table])

    def slices(list, slice):
        for i in range(0, len(list), slice):
            yield list[i : i + slice]

    def fetch_metadata(Meta_data_table_df):

        available_metadata = Meta_data_table_df.columns.tolist()[1:]
        win2_active = True
        meta_data_to_test_list = list(slices([sg.Radio(name, "metadata", key=name, default=True) for name in sorted(available_metadata)], 5))
        layout2 = [[sg.Text("Taxon list", size=(20,1))],
        [sg.Frame(layout = meta_data_to_test_list, title = 'Check metadata to test')],
        [sg.Text('',size=(1,1))],
        [sg.Button('Calculate')],
        [sg.Button('Skip')]]


        win2 = sg.Window('Taxon list', layout2, keep_on_top=True)

        while True:
            event2, values2 = win2.Read()
            if event2 == 'Calculate':
                for input_value, test in values2.items():
                    if (test == True):
                        meta_data_to_test = input_value
                win2.Close()
                win2_active = False
                return meta_data_to_test

            if event2 is None or event2 == "Skip":
                win2.Close()
                win2_active = False
                return False

    def gbif_requester(species_name):

        ## create an html session
        with requests_html.HTMLSession() as session:
            ## generate html request name
            request_name = '%20'.join(species_name.split(' '))

            ## request that name
            r = session.get('https://api.gbif.org/v1/species?name=%s&limit=1' % request_name)

            ## parse json
            res = json.loads(r.text)

            ## return the link
            if res['results']:
                species_id = res['results'][0]['key']
                link = ['https://www.gbif.org/species/%s' % res['results'][0]['key']]
            else:
                species_id = False
                link = [""]

            if species_id != False:
                ## collect vernacular name
                ## create an html session
                with requests_html.HTMLSession() as session:
                    ## request that name
                    r = session.get('https://api.gbif.org/v1/species/%s/vernacularNames' % species_id)

                    try:
                        ## parse json
                        res = json.loads(r.text)
                        ## search for english common name
                        name = list(set([name["vernacularName"] for name in res['results'] if name["language"] == "eng"]))
                        if len(name) > 1:
                            ## remove duplicates from the vernacular name list
                            name = list(set([i.lower() for i in name]))
                            name = [', '.join(name)]
                        if name == []:
                            name = [""]
                        return name + link
                    except:
                        return [""] + link
            else:
                return ["", ""]

    def pw_distance(seq1, seq2):
        len1 = len(seq1)
        len2 = len(seq2)
        length = max([len1, len2])
        alignments = pairwise2.align.globalxx(seq1, seq2)
        max_score = max([alignment.score for alignment in alignments])
        diff = 100 - (max_score / length * 100)
        diff = round(diff, 2)
        return(diff)

    # load the TaXon table
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('nan')
    TaXon_table_df_metadata = collect_metadata(TaXon_table_df)
    TaXon_table_df = strip_metadata(TaXon_table_df)
    samples = TaXon_table_df.columns.tolist()[10:]
    n_samples = len(samples)
    total_reads = sum([sum(reads) for reads in TaXon_table_df[samples].values.tolist()])

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")

    if use_metadata == True:
        if Meta_data_table_xlsx.exists():
            ## load metadata list
            Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0).fillna("nan")
            Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()
            meta_data_to_test = fetch_metadata(Meta_data_table_df)

            if meta_data_to_test != False:

                metadata_list = Meta_data_table_df[meta_data_to_test].values.tolist()
                metadata_loc = Meta_data_table_df.columns.tolist().index(meta_data_to_test)

                ## drop samples with metadata called nan (= empty)
                drop_samples = [i[0] for i in Meta_data_table_df.values.tolist() if i[metadata_loc] == "nan"]

                if drop_samples != []:
                    ## filter the TaXon table
                    TaXon_table_df = TaXon_table_df.drop(drop_samples, axis=1)
                    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
                    ## also remove empty OTUs
                    row_filter_list = []
                    for row in TaXon_table_df.values.tolist():
                        reads = set(row[10:])
                        if reads != {0}:
                            row_filter_list.append(row)
                    columns = TaXon_table_df.columns.tolist()
                    TaXon_table_df = pd.DataFrame(row_filter_list, columns=columns)
                    Meta_data_table_df = pd.DataFrame([i for i in Meta_data_table_df.values.tolist() if i[0] not in drop_samples], columns=Meta_data_table_df.columns.tolist())
                    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

                metadata_samples_dict = {}

                for i in Meta_data_table_df[["Samples", meta_data_to_test]].values.tolist():
                    metadata = i[1]
                    sample = i[0]
                    if metadata in metadata_samples_dict.keys():
                        metadata_samples_dict[metadata] = metadata_samples_dict[metadata] + [sample]
                    else:
                        metadata_samples_dict[metadata] = [sample]
                for metadata, samples in metadata_samples_dict.items():
                    TaXon_table_df[metadata] = [sum(i) for i in TaXon_table_df[samples].values.tolist()]
                    TaXon_table_df = TaXon_table_df.drop(samples, axis=1)
                samples = TaXon_table_df.columns.tolist()[10:]
                n_samples = len(samples)
                total_reads = sum([sum(reads) for reads in TaXon_table_df[samples].values.tolist()])
        else:
            sg.PopupOK("Warning: Metadata table does not exists.\n\nPlease first create a metadata table!\n\nSkipping metadata occupancy.", title="Warning")

    taxonomic_levels = TaXon_table_df.columns.values.tolist()[1:7]
    TaXon_table_df["FullName"] = ['<>'.join(t) for t in TaXon_table_df[taxonomic_levels].values.tolist()]

    ## create an empty data frame
    taxon_list_df = pd.DataFrame()
    ## add unique species IDs
    unique_species = list(set(['<>'.join(t) for t in TaXon_table_df[taxonomic_levels].values.tolist()]))
    taxon_list_df["FullName"] = unique_species

    # create the output files
    if taxon_list_output_file_name == '':
        taxon_list_output_file_name = "my_taxon_list"
    output_xlsx = Path(str(path_to_outdirs) + "/" + "Taxon_lists" + "/" + taxon_list_output_file_name + ".xlsx")

    answer = "Yes"
    if output_xlsx.exists():
        answer = sg.PopupYesNo("Taxon list already exists! Overwrite?")

    if answer == "Yes":
        ## create a species dict

        taxon_dict = {}
        for OTU in TaXon_table_df.values.tolist():
            taxon = OTU[-1]
            if taxon not in taxon_dict.keys():
                n_reads = sum(OTU[10:-1])
                n_OTUs = 1
                sequence = [OTU[9]]
                taxon_dict[taxon] = [n_reads, n_OTUs, sequence]
            else:
                n_reads = taxon_dict[taxon][0] + sum(OTU[10:-1])
                n_OTUs = taxon_dict[taxon][1] + 1
                sequence = taxon_dict[taxon][2] + [OTU[9]]
                taxon_dict[taxon] = [n_reads, n_OTUs, sequence]

        sg.Popup('TTT will be minimized and no progress bar will be shown to increase the performance.\nPlease refer to the console for progress updates.')

        ## create output lists
        n_OTUs_list = []
        n_reads_list = []
        n_reads_rel_list = []
        max_dist_list = []
        min_dist_list = []
        avg_dist_list = []
        rel_occurrence_list = []
        abs_occurrences_list = []
        vernacular_name_list = []
        gbif_link_list = []

        ## calculate all statistics per taxon
        for taxon in tqdm(taxon_list_df["FullName"], desc='Creating taxon list'):
            ## calculate statistics
            n_reads = taxon_dict[taxon][0]
            n_OTUs = taxon_dict[taxon][1]
            sequences = taxon_dict[taxon][2]
            n_reads_rel = round(n_reads / total_reads * 100, 3)
            species_name = taxon.split("<>")[-1]

            ## OCCURENCES
            if use_metadata == True:
                df = pd.DataFrame([i[10:-1] for i in TaXon_table_df.values.tolist() if taxon in i], columns=samples)
                sample_occurrences = []
                for sample in samples:
                    if sum(df[sample]) != 0:
                        sample_occurrences.append(sample)
                rel_occurences = round(len(sample_occurrences) / n_samples * 100, 3)

                abs_occurrences = []
                for sample in samples:
                    if sample in sample_occurrences:
                        abs_occurrences.append('x')
                    else:
                        abs_occurrences.append('')
            else:
                df = pd.DataFrame([i[10:-1] for i in TaXon_table_df.values.tolist() if taxon in i], columns=samples)
                sample_occurrences = []
                for sample in samples:
                    if sum(df[sample]) != 0:
                        sample_occurrences.append(sample)
                rel_occurences = round(len(sample_occurrences) / n_samples * 100, 3)

                abs_occurrences = []
                for sample in samples:
                    if sample in sample_occurrences:
                        abs_occurrences.append('x')
                    else:
                        abs_occurrences.append('')

            ## INTRASPECIFIC DISTANCE
            if calc_dist == True:
                if n_OTUs != 1 and species_name != "nan":
                    distances = []

                    for main_sequence in tqdm(sequences, leave=False, desc='Calculating distances for {}'.format(species_name)):
                        for sub_sequence in sequences:
                            d = pw_distance(main_sequence, sub_sequence)
                            if d != 0.0:
                                distances.append(d)
                    max_dist = max(distances)
                    min_dist = min(distances)
                    avg_dist = round(mean(distances), 2)
                    max_dist_list.append(max_dist)
                    min_dist_list.append(min_dist)
                    avg_dist_list.append(avg_dist)
                else:
                    max_dist_list.append('')
                    min_dist_list.append('')
                    avg_dist_list.append('')

            ## GBIF LINK
            if create_gbif_link == True:
                if species_name != "nan":
                    res = gbif_requester(species_name)
                    vernacular_name_list.append(res[0])
                    gbif_link_list.append(res[1])
                else:
                    vernacular_name_list.append('')
                    gbif_link_list.append('')

            ## append remaining values
            n_OTUs_list.append(n_OTUs)
            n_reads_list.append(n_reads)
            n_reads_rel_list.append(n_reads_rel)
            rel_occurrence_list.append(rel_occurences)
            abs_occurrences_list.append(abs_occurrences)

        taxon_list_df["Phylum"] = [i.split("<>")[0] if i.split("<>")[0] != "nan" else "" for i in taxon_list_df["FullName"].values.tolist()]
        taxon_list_df["Class"] = [i.split("<>")[1] if i.split("<>")[1] != "nan" else "" for i in taxon_list_df["FullName"].values.tolist()]
        taxon_list_df["Order"] = [i.split("<>")[2] if i.split("<>")[2] != "nan" else "" for i in taxon_list_df["FullName"].values.tolist()]
        taxon_list_df["Family"] = [i.split("<>")[3] if i.split("<>")[3] != "nan" else "" for i in taxon_list_df["FullName"].values.tolist()]
        taxon_list_df["Genus"] = [i.split("<>")[4] if i.split("<>")[4] != "nan" else "" for i in taxon_list_df["FullName"].values.tolist()]
        taxon_list_df["Species"] = [i.split("<>")[5] if i.split("<>")[5] != "nan" else "" for i in taxon_list_df["FullName"].values.tolist()]
        taxon_list_df.drop('FullName', axis='columns', inplace=True)
        taxon_list_df["OTUs"] = n_OTUs_list
        taxon_list_df["Reads"] = n_reads_list
        taxon_list_df["Reads (%)"] = n_reads_rel_list
        taxon_list_df["Occurrence (%)"] = rel_occurrence_list
        df_occurrences = pd.DataFrame(abs_occurrences_list, columns=samples)
        taxon_list_df = taxon_list_df.join(df_occurrences)
        if calc_dist == True:
            taxon_list_df["Max. dist."] = max_dist_list
            taxon_list_df["Min. dist."] = min_dist_list
            taxon_list_df["Avg. dist."] = avg_dist_list
        if create_gbif_link == True:
            taxon_list_df["Vernacular name"] = vernacular_name_list
            taxon_list_df["GBIF link"] = gbif_link_list


        ##############################################################################

        taxon_list_df.to_excel(output_xlsx, sheet_name = 'Taxa', index=False)

        answer = sg.PopupYesNo("Open taxon list?")
        if answer == "Yes":
            open_table(output_xlsx)

        closing_text = "Taxon list is found under:\n" + '/'.join(str(output_xlsx).split("/")[-4:])
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        from taxontabletools.create_log import ttt_log
        ttt_log("taxon list", "analysis", TaXon_table_xlsx.name, output_xlsx.name, "", path_to_outdirs)



from pathlib import Path
import PySimpleGUI as sg
import statsmodels.api as sm
import pandas as pd
import plotly.express as px
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import scipy.stats, math
import numpy as np
from scipy import stats
from statistics import mean
from scipy.spatial import distance
import plotly.figure_factory as ff
from statistics import mean
import webbrowser, os
from taxontabletools.taxontable_manipulation import strip_metadata

def tc_alpha_diversity(TaXon_table_xlsx_1, TaXon_table_xlsx_2, path_to_outdirs, name_1, name_2, width_value, height_value, template, taxonomic_level, font_size, color_discrete_sequence):

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = "# " + taxonomic_level
        taxonomic_level = "ID"
    if taxonomic_level == "Phylum":
        taxon_title = "# phyla"
    if taxonomic_level == "Class":
        taxon_title = "# classes"
    if taxonomic_level == "Order":
        taxon_title = "# orders"
    if taxonomic_level == "Family":
        taxon_title = "# families"
    if taxonomic_level == "Genus":
        taxon_title = "# genera"
    if taxonomic_level == "Species":
        taxon_title = "# species"

    ## load dataframes and sample names
    df1 = pd.read_excel(TaXon_table_xlsx_1).fillna('nan')
    df1 = strip_metadata(df1)
    df1_samples = df1.columns.tolist()[10:]
    df2 = pd.read_excel(TaXon_table_xlsx_2).fillna('nan')
    df2 = strip_metadata(df2)
    df2_samples = df2.columns.tolist()[10:]

    ## count overall number of taxa
    df_1_taxa = len(list(set([i for i in df1[taxonomic_level].values.tolist() if i != "nan"])))
    df_2_taxa = len(list(set([i for i in df2[taxonomic_level].values.tolist() if i != "nan"])))

    ## count the number of taxa per sample for both tables
    df1_dict, df2_dict = {}, {}

    for sample in df1_samples:
        n_taxa = set([i[0] for i in df1[[taxonomic_level, sample]].values.tolist() if i[0] != "nan" and i[1] != 0])
        df1_dict[sample] = len(list(n_taxa))

    for sample in df2_samples:
        n_taxa = set([i[0] for i in df2[[taxonomic_level, sample]].values.tolist() if i[0] != "nan" and i[1] != 0])
        df2_dict[sample] = len(list(n_taxa))

    ## collect values
    y1 = list(df1_dict.values())
    y2 = list(df2_dict.values())
    text_values1 = list(df1_dict.keys())
    text_values2 = list(df2_dict.keys())

    ## create a boxlplot
    fig = go.Figure()
    fig.add_trace(go.Box(y=y1, name=name_1, text=text_values1, marker_color=color_discrete_sequence[0]))
    fig.add_trace(go.Box(y=y2, name=name_2, text=text_values2, marker_color=color_discrete_sequence[1]))
    fig.update_yaxes(title=taxon_title)
    fig.update_traces(boxpoints='all', jitter=0.5)
    fig.update_yaxes(rangemode="tozero")
    fig.update_layout(width=int(width_value), height=int(height_value), template=template, showlegend=False, font_size=font_size)

    ## create a folder if neccessary
    dirName = Path(str(path_to_outdirs) + "/Table_comparison/" + name_1 + "_" + name_2 + "/")
    if not os.path.exists(dirName):
        os.mkdir(Path(dirName))

    ## write ouput files
    output_pdf = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_alpha_diversity.pdf")
    output_html = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_alpha_diversity.html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = "Plots are found under: " + str(path_to_outdirs) + "/Table_comparison/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write to log
    from taxontabletools.create_log import ttt_log
    out = name_1 + ";" + name_2
    ttt_log("Table comparison alpha diversity", "analysis", out, output_pdf.name, "", path_to_outdirs)

def tc_overall_taxon_richness(TaXon_table_xlsx_1, TaXon_table_xlsx_2, path_to_outdirs, name_1, name_2, width_value, height_value, template, taxonomic_level, taxonomic_level_2, font_size, color_discrete_sequence):

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = "# " + taxonomic_level
        taxonomic_level = "ID"
    if taxonomic_level == "Phylum":
        taxon_title = "# phyla"
    if taxonomic_level == "Class":
        taxon_title = "# classes"
    if taxonomic_level == "Order":
        taxon_title = "# orders"
    if taxonomic_level == "Family":
        taxon_title = "# families"
    if taxonomic_level == "Genus":
        taxon_title = "# genera"
    if taxonomic_level == "Species":
        taxon_title = "# species"

    ## load dataframes and sample names
    df1 = pd.read_excel(TaXon_table_xlsx_1).fillna('nan')
    df1 = strip_metadata(df1)
    df1_samples = df1.columns.tolist()[10:]
    df2 = pd.read_excel(TaXon_table_xlsx_2).fillna('nan')
    df2 = strip_metadata(df2)
    df2_samples = df2.columns.tolist()[10:]

    ## collect all taxa for the second (i.e. higher taxonomic level)
    taxa2_list = sorted(list(set([i for i in df1[taxonomic_level_2].values.tolist() + df2[taxonomic_level_2].values.tolist() if i != "nan"])))

    ## store the number of taxa per taxon (e.g. species per phylum)
    taxa_dict_1, taxa_dict_2 = {}, {}

    ## loop through taxa
    for taxon in taxa2_list:
        taxa_dict_1[taxon] = len(set([i[0] for i in df1[[taxonomic_level, taxonomic_level_2]].values.tolist() if i[0] != "nan" and i[1] == taxon]))
        taxa_dict_2[taxon] = len(set([i[0] for i in df2[[taxonomic_level, taxonomic_level_2]].values.tolist() if i[0] != "nan" and i[1] == taxon]))

    ## collect values
    x = list(taxa_dict_1.values())
    y = list(taxa_dict_2.values())
    text_values = list(taxa_dict_2.keys())

    ## calcuate spearman correlation
    spearman = scipy.stats.spearmanr(x, y)
    spearman_p = spearman[1]

    if spearman_p <= 0.05:
        spearman_rho = str(round(spearman[0], 3)) + "*"
    else:
        spearman_rho = str(round(spearman[0], 3))

    ## plot data
    fig = go.Figure()
    ## add scatter plot
    fig.add_trace(go.Scatter(name="", x=x, y=y, mode='markers', text=text_values, marker=dict(color=color_discrete_sequence[0], size=15)))
    ## add best fit
    bestfit = sm.OLS(y,sm.add_constant(x)).fit().fittedvalues
    fig.add_trace(go.Scatter(name="rho=" + spearman_rho, x=x, y=bestfit, mode='lines', opacity=0.2, marker=dict(color="black")))
    fig.update_yaxes(title=taxon_title + " (" + name_2 + ")")
    fig.update_xaxes(title=taxon_title + " (" + name_1 + ")")
    title_text = taxonomic_level + " per " + taxonomic_level_2.lower() + " (rho=" + spearman_rho + ")"
    fig.update_layout(title=title_text, width=int(width_value), height=int(height_value), template=template, showlegend=False, font_size=font_size)

    ## create a folder if neccessary
    dirName = Path(str(path_to_outdirs) + "/Table_comparison/" + name_1 + "_" + name_2 + "/")
    if not os.path.exists(dirName):
        os.mkdir(Path(dirName))

    ## write ouput files
    output_pdf = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "per" + taxonomic_level_2 + "_taxonomic_richness_scatter.pdf")
    output_html = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "per" + taxonomic_level_2 + "_taxonomic_richness_scatter.html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = "Plots are found under: " + str(path_to_outdirs) + "/Table_comparison/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write to log
    from taxontabletools.create_log import ttt_log
    out = name_1 + ";" + name_2
    ttt_log("Table comparison taxonomic richness scatter", "analysis", out, output_pdf.name, "", path_to_outdirs)

def tc_seperate_taxon_richness(TaXon_table_xlsx_1, TaXon_table_xlsx_2, path_to_outdirs, name_1, name_2, width_value, height_value, template, taxonomic_level, taxonomic_level_2, font_size, color_discrete_sequence, min_show):

    ## load dataframes and sample names
    df1 = pd.read_excel(TaXon_table_xlsx_1).fillna('nan')
    df1 = strip_metadata(df1)
    df1_samples = df1.columns.tolist()[10:]
    df2 = pd.read_excel(TaXon_table_xlsx_2).fillna('nan')
    df2 = strip_metadata(df2)
    df2_samples = df2.columns.tolist()[10:]

    ## collect all taxa for the second (i.e. higher taxonomic level)
    taxa2_list = sorted(list(set([i for i in df1[taxonomic_level_2].values.tolist() + df2[taxonomic_level_2].values.tolist() if i != "nan"])))

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = "# " + taxonomic_level
        taxonomic_level = "ID"
    if taxonomic_level == "Phylum":
        taxon_title = "# phyla"
    if taxonomic_level == "Class":
        taxon_title = "# classes"
    if taxonomic_level == "Order":
        taxon_title = "# orders"
    if taxonomic_level == "Family":
        taxon_title = "# families"
    if taxonomic_level == "Genus":
        taxon_title = "# genera"
    if taxonomic_level == "Species":
        taxon_title = "# species"

    non_shown = []
    y1_dict = {}
    y2_dict = {}


    ## loop through taxa
    for taxon in taxa2_list:

        y1, y2, text_values1, text_values2 = [], [], [], []
        min_show = int(min_show)

        for sample in df1_samples:
            n1 = len(set([i[0] for i in df1[[taxonomic_level, taxonomic_level_2, sample]].values.tolist() if i[0] != "nan" and i[1] == taxon and i[2] != 0]))
            y1.append(n1)
            text_values1.append(sample)

        for sample in df2_samples:
            n2 = len(set([i[0] for i in df2[[taxonomic_level, taxonomic_level_2, sample]].values.tolist() if i[0] != "nan" and i[1] == taxon and i[2] != 0]))
            y2.append(n2)
            text_values2.append(sample)

        if max(y1+y2) >= min_show:
            y1_dict[taxon] = [y1, text_values1]
            y2_dict[taxon] = [y2, text_values2]
        else:
            non_shown.append(taxon)

    ## plot
    ## create a boxlplot
    n_cols = 2
    n_rows = math.ceil(len(y1_dict.keys()) / n_cols)
    fig = make_subplots(rows=n_rows, cols=n_cols, vertical_spacing = 0.1, horizontal_spacing = 0.1, subplot_titles=list(y1_dict.keys()))
    col_counter = 1
    row_counter = 1

    for taxon in y1_dict.keys():
        fig.add_trace(go.Box(y=y1_dict[taxon][0], name=name_1, text=y1_dict[taxon][1], marker_color=color_discrete_sequence[0], marker_line_color="Black", marker_line_width=0.2), col=col_counter, row=row_counter)
        fig.add_trace(go.Box(y=y2_dict[taxon][0], name=name_2, text=y2_dict[taxon][1], marker_color=color_discrete_sequence[1], marker_line_color="Black", marker_line_width=0.2), col=col_counter, row=row_counter)
        fig.update_yaxes(rangemode="tozero", col=col_counter, row=row_counter)
        fig.update_traces(boxpoints='all', jitter=0.4, col=col_counter, row=row_counter)

        if col_counter == 1:
            fig.update_yaxes(title=taxon_title, col=col_counter, row=row_counter)

        ## iterate through rows and columns
        col_counter += 1
        if col_counter == n_cols+1:
            col_counter = 1
            row_counter += 1


    fig.update_layout(showlegend=False, width=int(width_value), height=int(height_value), template=template, font_size=font_size)

    ## create a folder if neccessary
    dirName = Path(str(path_to_outdirs) + "/Table_comparison/" + name_1 + "_" + name_2 + "/")
    if not os.path.exists(dirName):
        os.mkdir(Path(dirName))

    ## write ouput files
    output_pdf = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_per_" + taxonomic_level_2 + "_taxonomic_richness_boxplots.pdf")
    output_html = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_per_" + taxonomic_level_2 + "_taxonomic_richness_boxplots.html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = str(len(non_shown)) + " taxa fell below the treshold and are not shown.\n\nPlots are found under: " + str(path_to_outdirs) + "/Table_comparison/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write to log
    from taxontabletools.create_log import ttt_log
    out = name_1 + ";" + name_2
    ttt_log("Table comparison taxonomic richness boxplot", "analysis", out, output_pdf.name, "", path_to_outdirs)

def tc_pairwise_sample_comparison(TaXon_table_xlsx_1, TaXon_table_xlsx_2, path_to_outdirs, name_1, name_2, width_value, height_value, template, taxonomic_level, font_size, color_discrete_sequence):

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level
        taxonomic_level = "ID"
    if taxonomic_level == "Phylum":
        taxon_title = "phyla"
    if taxonomic_level == "Class":
        taxon_title = "classes"
    if taxonomic_level == "Order":
        taxon_title = "orders"
    if taxonomic_level == "Family":
        taxon_title = "families"
    if taxonomic_level == "Genus":
        taxon_title = "genera"
    if taxonomic_level == "Species":
        taxon_title = "species"

    ## load dataframes and sample names
    df1 = pd.read_excel(TaXon_table_xlsx_1).fillna('nan')
    df1 = strip_metadata(df1)
    df1_samples = df1.columns.tolist()[10:]
    df2 = pd.read_excel(TaXon_table_xlsx_2).fillna('nan')
    df2 = strip_metadata(df2)
    df2_samples = df2.columns.tolist()[10:]

    ## collect all taxa for the second (i.e. higher taxonomic level)
    taxa_blacklist = sorted(list(set([i for i in df1[taxonomic_level].values.tolist() + df2[taxonomic_level].values.tolist() if i != "nan"])))

    ## only proceed if both files have the same samples!
    if sorted(df1_samples) != sorted(df2_samples):
        sg.PopupError("Samples do not match between files!")
    else:

        jaccard_dict = {}
        jaccard_values = []

        for sample in df1_samples:

            n1 = set([i[0] for i in df1[[taxonomic_level, sample]].values.tolist() if i[0] != 'nan' and i[1] != 0])
            n2 = set([i[0] for i in df2[[taxonomic_level, sample]].values.tolist() if i[0] != 'nan' and i[1] != 0])

            col1 = []
            col2 = []
            for taxon in list(set(list(n1) + list(n2))):
                if taxon in n1:
                    col1.append(1)
                else:
                    col1.append(0)
                if taxon in n2:
                    col2.append(1)
                else:
                    col2.append(0)

            dist = round(distance.jaccard(col1, col2), 2)

            jaccard_dict[sample] = str(dist)
            jaccard_values.append(dist)

        avg_jaccard = round(mean(jaccard_values),2)
        min_jaccard = min(jaccard_values)
        max_jaccard = max(jaccard_values)

        jaccard_dict = dict(sorted(jaccard_dict.items(), key=lambda item: item[1]))

        fig = make_subplots(rows=1, cols=3, column_widths=[0.8, 0.2, 0.2], vertical_spacing = 0.05, horizontal_spacing = 0.05)

        y_table_1_list = []
        y_shared_list = []
        y_table_2_list = []
        x_samples = []
        y_n_taxa = []

        for sample in jaccard_dict.keys():

            table_1_list = []
            shared_list = []
            table_2_list = []

            n1 = set([i[0] for i in df1[[taxonomic_level, sample]].values.tolist() if i[1] != 0 and i[0] != "nan"])
            n2 = set([i[0] for i in df2[[taxonomic_level, sample]].values.tolist() if i[1] != 0 and i[0] != "nan"])

            n_taxa = len(set(list(n1) + list(n2)))

            if n_taxa != 0:
                if len(n1) != 0 and len(n2) != 0:
                    table_1_list.append(len(n1 - n2))
                    shared_list.append(len(n1 & n2))
                    table_2_list.append(len(n2 - n1))
                elif len(n1) == 0:
                    table_1_list.append(0)
                    shared_list.append(0)
                    table_2_list.append(len(n2))
                elif len(n2) == 0:
                    table_1_list.append(len(n1))
                    shared_list.append(0)
                    table_2_list.append(0)

            table_1_only = sum(table_1_list) / n_taxa * 100
            shared = sum(shared_list) / n_taxa * 100
            table_2_only = sum(table_2_list) / n_taxa * 100

            y_table_1_list.append(table_1_only)
            y_shared_list.append(shared)
            y_table_2_list.append(table_2_only)
            x_samples.append(sample)
            y_n_taxa.append(n_taxa)


        ## add the values used for the plot to a dataframe
        df_out = pd.DataFrame()
        df_out["Sample"] = x_samples
        df_out["Only "+ name_1] = y_table_1_list
        df_out["Shared"] = y_shared_list
        df_out["Only " + name_2] = y_table_2_list

        ## Shared / Only barchart
        fig.add_trace(go.Bar(name="Shared", orientation='h', y=x_samples, x=y_shared_list, marker_color="rgb(102,194,164)"), row=1, col=1)
        fig.add_trace(go.Bar(name=name_1, orientation='h', y=x_samples, x=y_table_1_list, marker_color="rgb(141,160,203)"), row=1, col=1)
        fig.add_trace(go.Bar(name=name_2, orientation='h', y=x_samples, x=y_table_2_list, marker_color="rgb(252,141,98)"), row=1, col=1)
        fig.update_layout(barmode='stack', showlegend=False, width=int(width_value), height=int(height_value), template=template, title="", font_size=font_size)
        fig.update_yaxes(tickmode = 'linear', showgrid=False, row=1, col=1)
        fig.update_xaxes(title=taxon_title + " (%)", showgrid=True, row=1, col=1)

        ## Jaccard plot
        y = list(jaccard_dict.keys())
        x = [float(i) for i in list(jaccard_dict.values())]
        df_out["Jaccard dissimilarity"] = x
        fig.add_trace(go.Bar(y=y, x=x, name="Jaccard", orientation='h', marker_color="lightgrey"), row=1, col=2)
        fig.update_yaxes(tickmode = 'linear', showticklabels=False, showgrid=False, row=1, col=2)
        fig.update_xaxes(title="jaccard diss.", showgrid=True, range=[0, 1], autorange=False, tick0=0, dtick=0.5, row=1, col=2)

        y = x_samples
        x = y_n_taxa
        df_out["# taxa"] = x
        fig.add_trace(go.Bar(y=y, x=x, name="taxa", orientation='h', marker_color="lightgrey"), row=1, col=3)
        fig.update_yaxes(tickmode = 'linear', showticklabels=False, showgrid=False, row=1, col=3)
        fig.update_xaxes(title="# " + taxon_title, showgrid=True, row=1, col=3)

        min_shared = round(min(y_shared_list),2)
        avg_shared = round(mean(y_shared_list),2)
        max_shared = round(max(y_shared_list),2)

        df_out = df_out.sort_values("Jaccard dissimilarity", ascending=False)

        ## create a folder if neccessary
        dirName = Path(str(path_to_outdirs) + "/Table_comparison/" + name_1 + "_" + name_2 + "/")
        if not os.path.exists(dirName):
            os.mkdir(Path(dirName))

        ## write ouput files
        output_pdf = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_overview_chart.pdf")
        output_html = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_overview_chart.html")
        output_xlsx = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_overview_chart.xlsx")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))
        df_out.to_excel(output_xlsx, index=False)

        ## ask to show file
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## print closing text
        closing_text = "Plots are found under: " + str(path_to_outdirs) + "/Table_comparison/"
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        ## write to log
        from taxontabletools.create_log import ttt_log
        out = name_1 + ";" + name_2
        ttt_log("Table comparison overview", "analysis", out, output_pdf.name, "", path_to_outdirs)

def tc_pairwise_taxon_comparison(TaXon_table_xlsx_1, TaXon_table_xlsx_2, path_to_outdirs, name_1, name_2, width_value, height_value, template, taxonomic_level, font_size, color_discrete_sequence, method):

    ## load dataframes and sample names
    df1 = pd.read_excel(TaXon_table_xlsx_1).fillna('nan')
    df1 = strip_metadata(df1)
    df1_samples = df1.columns.tolist()[10:]
    df2 = pd.read_excel(TaXon_table_xlsx_2).fillna('nan')
    df2 = strip_metadata(df2)
    df2_samples = df2.columns.tolist()[10:]

    ## collect all taxa for the second (i.e. higher taxonomic level)
    taxa_blacklist = sorted(list(set([i for i in df1[taxonomic_level].values.tolist() + df2[taxonomic_level].values.tolist() if i != "nan"])))

    levels = ["Phylum", "Class", "Order", "Family", "Genus", "Species"]

    ## only proceed if both files have the same samples!
    if sorted(df1_samples) != sorted(df2_samples):
        sg.PopupError("Samples do not match between files!")
    else:
        ## count taxa
        taxon_dict_x, taxon_dict_y = {}, {}

        ## unique taxon list for each sample
        sample_taxa_1, sample_taxa_2 = {}, {}
        n_taxa_1, n_taxa_2 = {}, {}

        ## store information for each sample in the two dicts
        for sample in df1_samples:
            ## collect taxa for each sample
            lst1 = [i[:-1] for i in df1[levels + [sample]].values.tolist() if i[-1] != 0]
            lst2 = [i[:-1] for i in df2[levels + [sample]].values.tolist() if i[-1] != 0]

            ## remove duplicates
            lst1 = [list(x) for x in set(tuple(x) for x in lst1)]
            lst2 = [list(x) for x in set(tuple(x) for x in lst2)]

            ## count overall number of taxa for rel. abundance
            n_taxa_1[sample] = len(lst1)
            n_taxa_2[sample] = len(lst2)

            ## flatten list
            sample_taxa_1[sample] = [item for sublist in lst1 for item in sublist]
            sample_taxa_2[sample] = [item for sublist in lst2 for item in sublist]

        n_significant, values_dict = [], {}

        for taxon in taxa_blacklist:
            ## store x and y values for every taxon
            x_values = []
            y_values = []

            if method == "Relative":
                ## collect information for each sample
                for sample in df1_samples:
                    x_values.append(sample_taxa_1[sample].count(taxon) / n_taxa_1[sample] * 100)
                    y_values.append(sample_taxa_2[sample].count(taxon) / n_taxa_2[sample] * 100)
            else:
                ## collect information for each sample
                for sample in df1_samples:
                    x_values.append(sample_taxa_1[sample].count(taxon))
                    y_values.append(sample_taxa_2[sample].count(taxon))

            ## test for spearman correlation
            spearman = scipy.stats.spearmanr(x_values, y_values)
            if spearman[1] <= 0.05:
                n_significant.append(1)
                R = " (rho=" + str(round(spearman[0], 2)) + "*)"
                values_dict[taxon] = [taxon + R], x_values, y_values, df1_samples
            elif spearman[1] > 0.05:
                n_significant.append(1)
                R = " (rho=" + str(round(spearman[0], 2)) + ")"
                values_dict[taxon] = [taxon + R], x_values, y_values, df1_samples

        ## create the plot
        columns = 3
        rows = math.ceil(len(n_significant) / columns)

        ## create a new figure
        titles = [i[0][0] for i in values_dict.values()]
        fig = make_subplots(rows=rows, cols=columns, subplot_titles=titles)

        col_count = 1
        row_count = 1
        for taxon in values_dict.keys():
            ## collect x values (aka df1 values)
            x = values_dict[taxon][1]
            ## collect x values (aka df2 values)
            y = values_dict[taxon][2]
            ## collect sample names
            text = values_dict[taxon][3]

            ## add scatter plot
            fig.add_trace(go.Scatter(name=taxon, x=x, y=y, mode='markers', text=text, marker_color="Blue"), row=row_count, col=col_count)

            ## add best fit
            bestfit = sm.OLS(y,sm.add_constant(x)).fit().fittedvalues
            fig.add_trace(go.Scatter(name="", x=x, y=bestfit, mode='lines', opacity=0.2, marker=dict(color="black")), row=row_count, col=col_count)

            ## add y axis titles
            if col_count == 1:
                if method == "Relative":
                    y_title = "taxa % (" + name_2 + ")"
                else:
                    y_title = "# taxa (" + name_2 + ")"
                fig.update_yaxes(title=y_title, row=row_count, col=col_count)

            ## add x axis titles
            if row_count == rows:
                if method == "Relative":
                    x_title = "taxa % (" + name_1 + ")"
                else:
                    x_title = "# taxa (" + name_1 + ")"
                fig.update_xaxes(title=x_title, row=row_count, col=col_count)

            ## increase column count
            col_count += 1
            ## increase row count inf max is reached and reset column count
            if col_count > columns:
                col_count = 1
                row_count += 1

        fig.update_layout(barmode='stack', showlegend=False, width=int(width_value), height=int(height_value), template=template, title="", font_size=font_size)

        ## create a folder if neccessary
        dirName = Path(str(path_to_outdirs) + "/Table_comparison/" + name_1 + "_" + name_2 + "/")
        if not os.path.exists(dirName):
            os.mkdir(Path(dirName))

        ## write ouput files
        output_pdf = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_overview_chart.pdf")
        output_html = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "_overview_chart.html")
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html))

        ## ask to show file
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## print closing text
        closing_text = "Plots are found under: " + str(path_to_outdirs) + "/Table_comparison/"
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        ## write to log
        from taxontabletools.create_log import ttt_log
        out = name_1 + ";" + name_2
        ttt_log("Table comparison overview", "analysis", out, output_pdf.name, "", path_to_outdirs)

def tc_taxon_occurrence(TaXon_table_xlsx_1, TaXon_table_xlsx_2, path_to_outdirs, name_1, name_2, width_value, height_value, template, taxonomic_level, taxonomic_level_2, font_size, color_discrete_sequence):

    ## load dataframes and sample names
    df1 = pd.read_excel(TaXon_table_xlsx_1).fillna('nan')
    df1 = strip_metadata(df1)
    df1_samples = df1.columns.tolist()[10:]
    df2 = pd.read_excel(TaXon_table_xlsx_2).fillna('nan')
    df2 = strip_metadata(df2)
    df2_samples = df2.columns.tolist()[10:]

    ## collect all taxa for the second (i.e. higher taxonomic level)
    taxa2_list = sorted(list(set([i for i in df1[taxonomic_level_2].values.tolist() + df2[taxonomic_level_2].values.tolist() if i != "nan"])))

    ## number of unique species
    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = "# " + taxonomic_level
        taxonomic_level = "ID"
    if taxonomic_level == "Phylum":
        taxon_title = "# unique phyla"
    if taxonomic_level == "Class":
        taxon_title = "# unique classes"
    if taxonomic_level == "Order":
        taxon_title = "# unique orders"
    if taxonomic_level == "Family":
        taxon_title = "# unique families"
    if taxonomic_level == "Genus":
        taxon_title = "# unique genera"
    if taxonomic_level == "Species":
        taxon_title = "# unique species"

    taxa_dict = {}
    ## loop through taxa
    for taxon in taxa2_list:
        n1 = set([i[0] for i in df1[[taxonomic_level, taxonomic_level_2]].values.tolist() if i[0] != "nan" and i[1] == taxon])
        n2 = set([i[0] for i in df2[[taxonomic_level, taxonomic_level_2]].values.tolist() if i[0] != "nan" and i[1] == taxon])

        only_n1 = len(n1 - n2)
        only_n2 = len(n2 - n1)

        taxa_dict[taxon] = [only_n1, 0 - only_n2]

    taxa_dict = {k: v for k, v in sorted(taxa_dict.items(), key=lambda item: item[1], reverse=True)}

    fig = go.Figure()
    y1 = [i[0] for i in taxa_dict.values()]
    y2 = [i[1] for i in taxa_dict.values()]
    x = list(taxa_dict.keys())
    fig.add_trace(go.Bar(x=x, y=y2, base=0, name=name_2, marker_color=color_discrete_sequence[1]))
    fig.add_trace(go.Bar(x=x, y=y1, base=0, name=name_1, marker_color=color_discrete_sequence[0]))
    fig.update_layout(showlegend=True, width=int(width_value), height=int(height_value), template=template, title="", font_size=font_size, barmode="stack")
    fig.update_xaxes(tickmode='linear', tickangle=-90)
    fig.update_yaxes(title=taxon_title)

    ## create a folder if neccessary
    dirName = Path(str(path_to_outdirs) + "/Table_comparison/" + name_1 + "_" + name_2 + "/")
    if not os.path.exists(dirName):
        os.mkdir(Path(dirName))

    ## write ouput files
    output_pdf = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "per" + taxonomic_level_2 + "_unique_taxa.pdf")
    output_html = Path(str(dirName) + "/" + name_1 + "_" + name_2 + "_" + taxonomic_level + "per" + taxonomic_level_2 + "_unique_taxa.html")
    fig.write_image(str(output_pdf))
    fig.write_html(str(output_html))

    ## ask to show file
    answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
    if answer == "Yes":
        webbrowser.open('file://' + str(output_html))

    ## print closing text
    closing_text = "Plots are found under: " + str(path_to_outdirs) + "/Table_comparison/"
    sg.Popup(closing_text, title="Finished", keep_on_top=True)

    ## write to log
    from taxontabletools.create_log import ttt_log
    out = name_1 + ";" + name_2
    ttt_log("Table comparison unique taxa", "analysis", out, output_pdf.name, "", path_to_outdirs)
















#



import os, webbrowser
import pandas as pd
from pandas import DataFrame
from pathlib import Path
import plotly.graph_objects as go
import PySimpleGUI as sg
import numpy as np
from plotly.subplots import make_subplots
from taxontabletools.taxontable_manipulation import strip_metadata
from taxontabletools.taxontable_manipulation import aggregate_taxontable

def site_occupancy_barchart(TaXon_table_xlsx, meta_data_to_test, taxonomic_level, path_to_outdirs, x_site_occ, y_site_occ, template, theme, font_size):

    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level
        taxonomic_level = "ID"

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('nan')
    TaXon_table_df = strip_metadata(TaXon_table_df)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header = 0).fillna("nan")
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    metadata_list = Meta_data_table_df[meta_data_to_test].values.tolist()
    metadata_loc = Meta_data_table_df.columns.tolist().index(meta_data_to_test)

    ## drop samples with metadata called nan (= empty)
    drop_samples = [i[0] for i in Meta_data_table_df.values.tolist() if i[metadata_loc] == "nan"]

    if drop_samples != []:
        ## filter the TaXon table
        TaXon_table_df = TaXon_table_df.drop(drop_samples, axis=1)
        TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
        ## also remove empty OTUs
        row_filter_list = []
        for row in TaXon_table_df.values.tolist():
            reads = set(row[10:])
            if reads != {0}:
                row_filter_list.append(row)
        columns = TaXon_table_df.columns.tolist()
        TaXon_table_df = pd.DataFrame(row_filter_list, columns=columns)
        Meta_data_table_df = pd.DataFrame([i for i in Meta_data_table_df.values.tolist() if i[0] not in drop_samples], columns=Meta_data_table_df.columns.tolist())
        Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    TaXon_table_n_samples = len(TaXon_table_samples)
    n_sites = len(set(Meta_data_table_df[meta_data_to_test].tolist()))

    answer = "Ask"
    output_message = "No"

    if (sorted(TaXon_table_samples) == sorted(Meta_data_table_samples) and TaXon_table_n_samples != n_sites):

        site_occupancy_dict = {}

        sites = set(Meta_data_table_df[meta_data_to_test].tolist())

        for site in sites:
            # this can either be a species name or the above specified taxonomic level
            present_OTU_list = []

            # extract samples that belong to the site from the metadata file
            included_samples_list = Meta_data_table_df[Meta_data_table_df.values  == site]['Samples'].values.tolist()

            # count the number of samples per site to calculate the site occupancy
            n_samples = len(included_samples_list)

            # create a list of all species (or the specified taxonomic level)
            if taxonomic_level == "OTUs":
                taxonomic_level = "ID"
            overall_included_species_list = TaXon_table_df[taxonomic_level].values.tolist()
            # make the list unique
            overall_included_species_set = set(overall_included_species_list)
            # remove potential 'nan's from the list
            overall_included_species_set = [x for x in overall_included_species_set if str(x) != 'nan']

            # create a set of species that is present at the sites
            for sample in included_samples_list:

                OTUs_per_species_list = []

                # check the read abundaces for each sample
                read_abundace_list = TaXon_table_df[sample].values.tolist()

                # enumerate the read abundaces for each sample and collect all lines that have more than one read
                for i, read_abundance in enumerate(read_abundace_list):
                    species = TaXon_table_df[taxonomic_level][i]
                    # if reads are present, collect the species name (or the specified taxonomic level) from the TaXon table
                    if read_abundance != 0:
                        OTUs_per_species_list.append(species)

                # remove all nans
                OTUs_per_species_list = [x for x in OTUs_per_species_list if str(x) != 'nan']
                # make list unique
                OTUs_per_species_list = list(set(OTUs_per_species_list))
                # append to list of species for the current site
                present_OTU_list.append(OTUs_per_species_list)

            # flatten the list of present species per site
            present_OTU_list_flattened = [val for sublist in present_OTU_list for val in sublist]

            # store occupancy of each species in a dict, will be accessed by position in list
            occupancy_dict = {}

            # count the number of occurences for each species and calculate the occpancy based on the number of samples
            for species in overall_included_species_set:
                count = present_OTU_list_flattened.count(species)
                occupancy = count / n_samples * 100
                occupancy_dict[species] = occupancy

            occupancy_dict = {k: v for k, v in sorted(occupancy_dict.items(), key=lambda item: item[1])}
            occupancy_list = list(occupancy_dict.values())
            species_list = list(occupancy_dict.keys())

            if (taxonomic_level == "Species" or taxonomic_level == "Genus"):
                x_values = ["<i>" + taxon + "</i>" for taxon in species_list]
            else:
                x_values = species_list

            occupancy_plot_directory = Path(str(path_to_outdirs) + "/" + "Site_occupancy_plots" + "/" + TaXon_table_xlsx.stem)
            if not os.path.exists(occupancy_plot_directory):
                os.mkdir(occupancy_plot_directory)

            fig = go.Figure(data=[go.Bar(x=x_values, y=occupancy_list)])
            fig.update_traces(marker_color=color1, marker_line_color=color2,marker_line_width=0.6, opacity=opacity_value)
            fig.update_layout(title_text=site + " (" + taxonomic_level + ")", yaxis_title="occupancy (%)")
            fig.update_layout(height=int(y_site_occ), width=int(x_site_occ), template=template, font_size=font_size, title_font_size=font_size)
            fig.update_yaxes(range=[0,100])
            fig.update_xaxes(tickmode='linear')
            fig.update_xaxes(tickangle=-90)


            output_pdf = Path(str(occupancy_plot_directory) + "/" + site + "_" + taxonomic_level + ".pdf")
            output_html = Path(str(occupancy_plot_directory) + "/" + site + "_" + taxonomic_level + ".html")
            occupancy_table = Path(str(occupancy_plot_directory) + "/" + site + "_" + taxonomic_level + ".xlsx")
            fig.write_image(str(output_pdf))
            fig.write_html(str(output_html))
            occupancy_df = pd.DataFrame(occupancy_list, species_list)
            occupancy_df.columns = ["Occupancy"]
            occupancy_df.index.name = "Taxon"
            occupancy_df = occupancy_df.sort_values("Occupancy")
            # sort the table numerical if OTUs were chosen
            if taxonomic_level == "ID":
                sort_list = []
                for OTU in occupancy_df.index.tolist():
                    sort_list.append(int(OTU.split("_")[1]))
                occupancy_df["sort"] = sort_list
                occupancy_df = occupancy_df.sort_values("sort")
                occupancy_df = occupancy_df.drop("sort", axis=1)
            occupancy_df.to_excel(occupancy_table)

        ## ask to show file
        answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html))

        ## print closing text
        closing_text = "Site occupancy plots are found under:\n" + '/'.join(str(output_pdf).split("/")[-4:])
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        ## write to log
        from taxontabletools.create_log import ttt_log
        placeholder = TaXon_table_xlsx.name + " (multiple site occupancy plots)"
        ttt_log("site occupancy", "analysis", TaXon_table_xlsx.name, placeholder, meta_data_to_test, path_to_outdirs)

    else:
        sg.PopupError("Please check your Metadata file and Taxon table file: The samples do not match or the metadata is unique for all samples!", keep_on_top=True)

def site_occupancy_heatmap_pa(TaXon_table_xlsx, path_to_outdirs, template, height, width, meta_data_to_test, taxonomic_level, font_size, color_discrete_sequence, add_categories_sum):

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('unidentified')
    TaXon_table_df = strip_metadata(TaXon_table_df)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0).fillna("nan")
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    ## drop samples with metadata called nan (= empty)
    drop_samples = [i[0] for i in Meta_data_table_df.values.tolist() if i[1] == "nan"]

    if drop_samples != []:
        ## filter the TaXon table
        TaXon_table_df = TaXon_table_df.drop(drop_samples, axis=1)
        TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
        ## also remove empty OTUs
        row_filter_list = []
        for row in TaXon_table_df.values.tolist():
            reads = set(row[10:])
            if reads != {0}:
                row_filter_list.append(row)
        columns = TaXon_table_df.columns.tolist()
        TaXon_table_df = pd.DataFrame(row_filter_list, columns=columns)
        Meta_data_table_df = pd.DataFrame([i for i in Meta_data_table_df.values.tolist() if i[0] not in drop_samples], columns=Meta_data_table_df.columns.tolist())
        Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    metadata_list = Meta_data_table_df[meta_data_to_test].values.tolist()

    ## create a y axis title text
    taxon_title = taxonomic_level

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level
        taxonomic_level = "ID"

    if len(set(metadata_list)) == 1:
        sg.PopupError("Please choose more than one meta data category.")
    else:
        if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

            ## define variables
            samples = TaXon_table_samples
            OTU_abundances_dict = {}
            samples_metadata_list = []

            ## extract the relevant data
            TaXon_table_df = TaXon_table_df[[taxonomic_level] + samples]
            ## define an aggregation function to combine multiple hit of one taxonimic level
            aggregation_functions = {}
            ## define samples functions
            for sample in samples:
                ## 'sum' will calculate the sum of p/a data
                aggregation_functions[sample] = 'sum'
            ## define taxon level function
            aggregation_functions[taxonomic_level] = 'first'
            ## create condensed dataframe
            TaXon_table_df = TaXon_table_df.groupby(TaXon_table_df[taxonomic_level]).aggregate(aggregation_functions)
            if 'unidentified' in TaXon_table_df.index:
                TaXon_table_df = TaXon_table_df.drop('unidentified')

            ## create a list of samples for each category
            category_dict = {}
            for sample, category in zip(Meta_data_table_samples, metadata_list):
                if category not in category_dict.keys():
                    category_dict[category] = [sample]
                else:
                    category_dict[category] = category_dict[category] + [sample]

            ## collect all available taxa
            taxa = TaXon_table_df[taxonomic_level].values.tolist()

            ## check if the respective species are present in the collections
            taxon_presence_dict = {}
            n_rows, row_heights = [], []

            color_discrete_sequence = color_discrete_sequence * len(category_dict.keys())

            if (taxonomic_level == "Species" or taxonomic_level == "Genus"):
                x_values = ["<i>" + taxon + "</i>" for taxon in taxa]
            else:
                x_values = taxa

            if add_categories_sum == True:
                for samples in category_dict.values():
                    row_heights.append(len(samples))
                row_heights.append(len(set(metadata_list)))
                fig = make_subplots(rows=len(set(metadata_list)) + 1, cols=1, shared_xaxes=True, vertical_spacing=0.05, row_heights=row_heights)
            else:
                for samples in category_dict.values():
                    row_heights.append(len(samples))
                fig = make_subplots(rows=len(set(metadata_list)), cols=1, shared_xaxes=True, vertical_spacing=0.05, row_heights=row_heights)

            row = 1
            for metadata, samples in category_dict.items():
                if type(samples) == "str":
                    samples = [samples]
                z_values = []
                for sample in samples:
                    reads = TaXon_table_df[sample].values.tolist()
                    z_values = z_values + [[1 if x > 0 else 0 for x in reads]]
                y_values = samples
                fig.add_trace(go.Heatmap(z=z_values, x=x_values, y=y_values, showscale=False, xgap=1, ygap=1, hoverongaps = False, colorscale=[[0, "White"], [1, color_discrete_sequence[row-1]]]), row=row, col=1)
                row += 1

            if add_categories_sum == True:
                z_values, y_values = [], []
                for metadata, samples in category_dict.items():
                    reads = [sum(reads) for reads in TaXon_table_df[samples].values.tolist()]
                    z_values = z_values + [[1 if x > 0 else 0 for x in reads]]
                    y_values.append(metadata)
                fig.add_trace(go.Heatmap(z=z_values[::-1], x=x_values, y=y_values[::-1], showscale=False, xgap=1, ygap=1, hoverongaps = False, colorscale=[[0, "White"], [1, "Grey"]]), row=row, col=1)
                row += 1

            fig.update_layout(width=int(width), height=int(height), template="seaborn", font_size=font_size, yaxis_nticks=5, title_font_size=font_size)
            fig.update_xaxes(tickmode='linear')
            fig.update_yaxes(tickmode='linear')
            fig.update_xaxes(tickangle=-90)

            occupancy_plot_directory = Path(str(path_to_outdirs) + "/" + "Site_occupancy_plots" + "/" + TaXon_table_xlsx.stem)
            if not os.path.exists(occupancy_plot_directory):
                os.mkdir(occupancy_plot_directory)

            ## define output files
            output_pdf = Path(str(occupancy_plot_directory) + "/" + taxonomic_level + "_" + meta_data_to_test + "_heatmap.pdf")
            output_html = Path(str(occupancy_plot_directory) + "/" + taxonomic_level + "_" + meta_data_to_test + "_heatmap.html")

            ## write output files
            fig.write_image(str(output_pdf))
            fig.write_html(str(output_html))

            ## ask to show file
            answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
            if answer == "Yes":
                webbrowser.open('file://' + str(output_html))

            ## print closing text
            closing_text = "Site occupancy heatmaps are found under:\n" + '/'.join(str(output_pdf).split("/")[-4:])
            sg.Popup(closing_text, title="Finished", keep_on_top=True)

            ## write to log
            from taxontabletools.create_log import ttt_log
            placeholder = TaXon_table_xlsx.name + " (multiple site occupancy plots)"
            ttt_log("site occupancy", "analysis", TaXon_table_xlsx.name, "", meta_data_to_test, path_to_outdirs)


        else:
            sg.Popup("The metadata table and taXon table are not matching!")

def site_occupancy_heatmap_reads(TaXon_table_xlsx, path_to_outdirs, template, height, width, meta_data_to_test, taxonomic_level, font_size, color_discrete_sequence, add_categories_sum):

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('')
    TaXon_table_df = strip_metadata(TaXon_table_df)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0).fillna("nan")
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    ## drop samples with metadata called nan (= empty)
    drop_samples = [i[0] for i in Meta_data_table_df.values.tolist() if i[1] == "nan"]

    if drop_samples != []:
        ## filter the TaXon table
        TaXon_table_df = TaXon_table_df.drop(drop_samples, axis=1)
        TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]
        ## also remove empty OTUs
        row_filter_list = []
        for row in TaXon_table_df.values.tolist():
            reads = set(row[10:])
            if reads != {0}:
                row_filter_list.append(row)
        columns = TaXon_table_df.columns.tolist()
        TaXon_table_df = pd.DataFrame(row_filter_list, columns=columns)
        Meta_data_table_df = pd.DataFrame([i for i in Meta_data_table_df.values.tolist() if i[0] not in drop_samples], columns=Meta_data_table_df.columns.tolist())
        Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    metadata_list = Meta_data_table_df[meta_data_to_test].values.tolist()

    ## create a y axis title text
    taxon_title = taxonomic_level

    ## adjust taxonomic level if neccessary
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level
        taxonomic_level = "ID"

    if len(set(metadata_list)) == 1:
        sg.PopupError("Please choose more than one meta data category.")
    else:
        if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):

            ## define variables
            samples = TaXon_table_samples
            OTU_abundances_dict = {}
            samples_metadata_list = []

            ## aggregate taxontable, sort, and extract relevant taxa
            TaXon_table_df = aggregate_taxontable(TaXon_table_df, taxonomic_level)
            TaXon_table_df = TaXon_table_df.sort_values(['Phylum', taxonomic_level], ascending=[True, True])
            TaXon_table_df = TaXon_table_df.loc[TaXon_table_df[taxonomic_level] != ''][[taxonomic_level] + samples]

            ## create a list of samples for each category
            category_dict = {}
            for sample, category in zip(Meta_data_table_samples, metadata_list):
                if category not in category_dict.keys():
                    category_dict[category] = [sample]
                else:
                    category_dict[category] = category_dict[category] + [sample]

            ## collect all available taxa
            taxa = TaXon_table_df[taxonomic_level].values.tolist()

            ## make a copy of the original df
            TaXon_table_df_copy = TaXon_table_df.copy(deep=True)

            ## convert table to log reads
            for col in samples:
                TaXon_table_df[col] = [np.log(i) if i !=0 else 0 for i in TaXon_table_df[col].values.tolist()]

            ## calculate log max
            global_max = max([max(i) for i in TaXon_table_df[samples].values.tolist()])

            ## check if the respective species are present in the collections
            taxon_presence_dict = {}
            n_rows, row_heights = [], []

            colorscales = ['blues', 'reds', 'greens', 'oranges' ,'BuPu', 'Greys'] * len(meta_data_to_test)

            if (taxonomic_level == "Species" or taxonomic_level == "Genus"):
                x_values = ["<i>" + taxon + "</i>" for taxon in taxa]
            else:
                x_values = taxa

            if add_categories_sum == True:
                for samples in category_dict.values():
                    row_heights.append(len(samples))
                row_heights.append(len(set(metadata_list)))
                fig = make_subplots(rows=len(set(metadata_list)) + 1, cols=1, shared_xaxes=True, vertical_spacing=0.05, row_heights=row_heights)
            else:
                for samples in category_dict.values():
                    row_heights.append(len(samples))
                fig = make_subplots(rows=len(set(metadata_list)), cols=1, shared_xaxes=True, vertical_spacing=0.05, row_heights=row_heights)

            row = 1
            for metadata, samples in category_dict.items():
                if type(samples) == "str":
                    samples = [samples]
                z_values = []
                for sample in samples:
                    reads = TaXon_table_df[sample].values.tolist()
                    z_values = z_values + [reads]
                y_values = samples
                fig.add_trace(go.Heatmap(z=z_values, x=x_values, y=y_values, showscale=False, xgap=1, ygap=1, hoverongaps = False, zmin=0, zmax=global_max, colorscale=colorscales[row-1]), row=row, col=1)
                row += 1

                fig.write_html("/Users/tillmacher/Desktop/Paper/eRNA_paper/TEST.html")

            if add_categories_sum == True:
                z_values, y_values = [], []
                for metadata, samples in category_dict.items():
                    reads = [sum(reads) for reads in TaXon_table_df_copy[samples].values.tolist()]
                    z_values = z_values + [[np.log(x) if x != 0 else 0 for x in reads]]
                    y_values.append(metadata)
                fig.add_trace(go.Heatmap(z=z_values[::-1], x=x_values, y=y_values[::-1], showscale=False, xgap=1, ygap=1, hoverongaps = False, colorscale='gray_r'), row=row, col=1)
                row += 1

            fig.update_layout(width=int(width), height=int(height), template="seaborn", font_size=font_size, yaxis_nticks=5, title_font_size=font_size)
            fig.update_xaxes(tickmode='linear')
            fig.update_yaxes(tickmode='linear')
            fig.update_xaxes(tickangle=-90)

            occupancy_plot_directory = Path(str(path_to_outdirs) + "/" + "Site_occupancy_plots" + "/" + TaXon_table_xlsx.stem)
            if not os.path.exists(occupancy_plot_directory):
                os.mkdir(occupancy_plot_directory)

            ## define output files
            output_pdf = Path(str(occupancy_plot_directory) + "/" + taxonomic_level + "_" + meta_data_to_test + "_heatmap_reads.pdf")
            output_html = Path(str(occupancy_plot_directory) + "/" + taxonomic_level + "_" + meta_data_to_test + "_heatmap_reads.html")

            ## write output files
            fig.write_image(str(output_pdf))
            fig.write_html(str(output_html))

            ## ask to show file
            answer = sg.PopupYesNo('Show plot?', keep_on_top=True)
            if answer == "Yes":
                webbrowser.open('file://' + str(output_html))

            ## print closing text
            closing_text = "Site occupancy heatmaps are found under:\n" + '/'.join(str(output_pdf).split("/")[-4:])
            sg.Popup(closing_text, title="Finished", keep_on_top=True)

            ## write to log
            from taxontabletools.create_log import ttt_log
            placeholder = TaXon_table_xlsx.name + " (multiple site occupancy plots)"
            ttt_log("site occupancy", "analysis", TaXon_table_xlsx.name, "", meta_data_to_test, path_to_outdirs)


        else:
            sg.Popup("The metadata table and taXon table are not matching!")


import os
import pandas as pd
import numpy as np
from datetime import datetime
from pathlib import Path
import networkx as nx
import matplotlib.pyplot as plt
import PySimpleGUI as sg

def ttt_log(task, type, input_file, output_file, parameters, path_to_outdirs):

    dateTimeObj = datetime.now()
    time = str(dateTimeObj)

    log_xlsx = Path(str(path_to_outdirs) + "/log.xlsx")

    if os.path.exists(Path(log_xlsx)):
        log_df = pd.read_excel(log_xlsx)
        log_df = pd.DataFrame(log_df.values.tolist() + [[task, input_file, output_file, parameters, time, type]],columns=log_df.columns.tolist())
    else:
        log_df = pd.DataFrame([[task, input_file, output_file, parameters, time, type]],columns=["task", "input", "output", "parameters", "time", "type"])

    log_df.to_excel(log_xlsx, index=False)

def ttt_log_network(path_to_outdirs):

    log_xlsx = Path(str(path_to_outdirs) + "/log.xlsx")

    if not os.path.exists(Path(log_xlsx)):
        sg.PopupError("No log file exists!", title="Error", keep_on_top=True)
        raise RuntimeError

    win2_layout = [
                [sg.Text('',size=(1,1))],
                [sg.Text('Please adjust the plot size to the size of the network.')],
                [sg.Text("Plot size (w,h):"), sg.Input("12", size=(3,1), key="x_log"), sg.Input("12", size=(3,1), key="y_log")],
                [sg.Text('',size=(1,1))],
                [sg.Button('Continue', key='Continue')],
                ]

    win2 = sg.Window('Projects', win2_layout)
    event, values = win2.read()
    x_log = int(values["x_log"])
    y_log = int(values["y_log"])
    win2.close()

    df = pd.read_excel(log_xlsx)

    edge_label_dict = {}
    node_label_dict = {}
    for task in df.values.tolist():
        # collect data from the log table
        input = task[1]
        output = task[2]
        edge_label = task[0]
        type = task[-1]
        # create a dict entry for the edge name
        edge_label_dict[(input, output)] = edge_label
        # createa a dict entry for the node name
        if type == "processing":
            node_label_dict[input] = input
            node_label_dict[output] = output
        else:
            node_label_dict[input] = input
            node_label_dict[output] = ""

    plt.figure(figsize=(x_log, y_log))
    G=nx.from_pandas_edgelist(df, 'input', 'output', create_using=nx.Graph())
    nx.draw(G, with_labels=False, node_color="skyblue", node_size=50, width=3, edge_color = "skyblue", edge_cmap=plt.cm.Blues, pos=nx.kamada_kawai_layout(G), font_size=8)
    nx.draw_networkx_edge_labels(G, pos=nx.kamada_kawai_layout(G), edge_labels=edge_label_dict, font_color='blue', font_size=6)
    nx.draw_networkx_labels(G, pos=nx.kamada_kawai_layout(G), labels=node_label_dict, font_color='black', font_size=8)
    plt.axis('off')
    plt.show(block=False)

    answer = sg.PopupYesNo('Save figure?', keep_on_top=True)
    if answer == "Yes":
        log_pdf = Path(str(path_to_outdirs) + "/log.pdf")
        plt.savefig(log_pdf)
        plt.close()
        closing_text = "Log network is found under:\n" + log_pdf.name
        sg.Popup(closing_text, title="Finished", keep_on_top=True)
    else:
        plt.close()



import pandas as pd
import itertools

def reduce_taxontable(df, taxonomic_level, samples, meta_data_to_test):

    if meta_data_to_test == '':
        ## extract the relevant data
        df = df[[taxonomic_level] + samples]
        ## define an aggregation function to combine multiple hit of one taxonimic level
        aggregation_functions = {}
        ## define samples functions
        for sample in samples:
            ## 'sum' will calculate the sum of p/a data
            aggregation_functions[sample] = 'sum'
        ## define taxon level function
        aggregation_functions[taxonomic_level] = 'first'
        ## create condensed dataframe
        df = df.groupby(df[taxonomic_level]).aggregate(aggregation_functions)
        if '' in df.index:
            df = df.drop('')

        return df

    else:
        ## collect metadata for each taxon
        taxon_metadata_dict = {}
        for row in df[[taxonomic_level, meta_data_to_test]].values.tolist():
            taxon = row[0]
            metadata = row[1]
            taxon_metadata_dict[taxon] = metadata

        ## extract the relevant data
        df = df[[taxonomic_level] + samples]
        ## define an aggregation function to combine multiple hit of one taxonimic level
        aggregation_functions = {}
        ## define samples functions
        for sample in samples:
            ## 'sum' will calculate the sum of p/a data
            aggregation_functions[sample] = 'sum'
        ## define taxon level function
        aggregation_functions[taxonomic_level] = 'first'
        ## create condensed dataframe
        df = df.groupby(df[taxonomic_level]).aggregate(aggregation_functions)
        if '' in df.index:
            df = df.drop('')

        ## add metadata back to taxa
        metadata_col = []
        for taxon in df[taxonomic_level].values.tolist():
            metadata_col.append(taxon_metadata_dict[taxon])
        df[meta_data_to_test] = metadata_col

        return df

def strip_metadata(df):
    " Strip the additional metadata from the dataframe to produce a clean TaxonTable "

    ## position of seq == acts as separator
    seq_loc = df.columns.get_loc("seq")

    ## samples are always assending the seq column
    samples = df.columns.tolist()[seq_loc+1:]

    ## standard columns
    standard_columns = ['ID', 'Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species', 'Similarity', 'Status', 'seq']

    ## extract legacy TTT frame
    df = df[standard_columns + samples]

    return df

def collect_metadata(df):
    " Collect the additional metadata from the dataframe to be able to place it back later "

    ## position of status and seq == acts as separator
    seq_loc = df.columns.get_loc("seq")
    status = df.columns.get_loc("Status")

    ## metadata is always located between status and seq
    metadata = df.columns.tolist()[status+1:seq_loc]

    ## collect IDs and metadata
    df = df[['ID'] + metadata]

    return df

def add_metadata(df, metadata_df):
    " Collect the additional metadata from the dataframe to add them back to a df "

    for metadata in  metadata_df.columns.tolist()[1:]:
        status_loc = df.columns.get_loc("Status") + 1
        ## loope through IDs and find all that still remain in the altered Taxon table (e.g. after filtering less OTUs might be present)
        sub_df = pd.DataFrame([i for i in metadata_df[['ID', metadata]].values.tolist() if i[0] in df['ID'].values.tolist()], columns = ['ID', metadata])
        ## insert sub_df containing the respective metadata after the Status column
        df.insert(status_loc, str(metadata), sub_df[metadata].values.tolist(), True)

    return df

def aggregate_taxontable(df, taxonomic_level):

    seq_loc = df.columns.get_loc("seq")
    samples = df.columns.tolist()[seq_loc+1:]
    taxonomic_levels = ['Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species']
    index = taxonomic_levels.index(taxonomic_level)+1
    taxonomic_levels = taxonomic_levels[:index]

    ## collect unique taxa
    taxa = df[taxonomic_levels].values.tolist()
    taxa.sort()
    unique_taxa = list(taxa for taxa, _ in itertools.groupby(taxa))

    df_list = []

    ## summarize reads for all taxa
    for taxon in unique_taxa:
        ## calculate the sum of reads
        reads = list(pd.DataFrame(
            [i[index:] for i in df[taxonomic_levels + samples].values.tolist() if i[0:index] == taxon]).sum())
        df_list.append(taxon + reads)

    df_out = pd.DataFrame(df_list, columns=taxonomic_levels + samples)

    return df_out



import PySimpleGUI as sg
import pandas as pd
import numpy as np
from pathlib import Path
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import itertools, webbrowser
from taxontabletools.taxontable_manipulation import strip_metadata

def per_taxon_analysis(TaXon_table_xlsx, height, width, taxonomic_level, path_to_outdirs, template, theme, font_size, clustering_unit):

    ## save the taxon title
    taxon_title = clustering_unit

    ## collect plot variables
    color1 = theme[0]
    color2 = theme[1]
    opacity_value = theme[2]
    height = int(height)
    width = int(width)

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('nan')
    TaXon_table_df = strip_metadata(TaXon_table_df)

    ## collect the taxa to test on
    taxa = sorted(list(set([taxon for taxon in TaXon_table_df[taxonomic_level].values.tolist() if taxon != "nan"])))

    ## check if there are more than 8 taxa
    answer = "Yes"
    if len(taxa) > 8:
        answer = sg.PopupYesNo("There are more than 8 taxa detected. This can render the plot difficult to read. Continue anyway?")

    if answer == "Yes":
        ## collect the OTUs
        OTUs = TaXon_table_df["ID"].values.tolist()

        ## count the number of OTUs per taxon
        n_OTUs = {taxon:TaXon_table_df[taxonomic_level].values.tolist().count(taxon) for taxon in taxa}
        n_OTUs = dict(sorted(n_OTUs.items(), key=lambda item: item[1], reverse=True))
        taxa = list(n_OTUs.keys())

        ## collect all OTUs on species level
        OTU_species = [OTU for OTU in TaXon_table_df[[taxonomic_level, "Species"]].values.tolist() if OTU[1] != "nan"]
        OTU_species.sort()
        OTU_species = list(k for k,_ in itertools.groupby(OTU_species))
        OTU_species = [OTU[0] for OTU in OTU_species]
        n_species = {taxon:OTU_species.count(taxon) for taxon in taxa}

        ## count reads for each taxon
        n_reads = []
        for taxon in taxa:
            n_reads.append(sum([sum(OTU[10:]) for OTU in TaXon_table_df[TaXon_table_df[taxonomic_level]==taxon].values.tolist()]))
        if (taxonomic_level == "Species" or taxonomic_level == "Genus"):
            x_values = ["<i>" + taxon + "</i>" for taxon in taxa]
        else:
            x_values = taxa

        ## calculate the read proportions
        reads_sum = sum(n_reads)
        n_reads = [round(reads / reads_sum * 100, 2) for reads in n_reads]

        ## create subplots
        fig = go.Figure()
        ## percentage of reads per taxonomic level
        hovertext = 'Taxon: %{x}, Reads: %{y}'
        fig.add_trace(go.Bar(hovertemplate=hovertext, name="",x=x_values, y=n_reads))
        fig.update_yaxes(title_text = "reads (%)", title_standoff=5)
        fig.update_traces(marker_color=color1, marker_line_color=color2, marker_line_width=1, opacity=opacity_value, showlegend=False)

        ## update the layout
        fig.update_layout(barmode='stack', uniformtext_mode='hide', height=int(height), width=int(width), template=template, showlegend=False, font_size=font_size, title_font_size=font_size)
        fig.update_xaxes(tickmode='linear')
        fig.update_xaxes(tickangle=-90)

        ## write ouput files
        output_pdf = Path(path_to_outdirs).joinpath('Per_taxon_statistics', '{}_{}_reads.pdf'.format(TaXon_table_xlsx.stem, taxonomic_level))
        output_html1 = Path(path_to_outdirs).joinpath('Per_taxon_statistics', '{}_{}_reads.html'.format(TaXon_table_xlsx.stem, taxonomic_level))
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html1))


        ## create subplots
        fig = go.Figure()
        ## Number of OTUs
        hovertext = 'Taxon: %{x}, OTUs: %{y}'
        title_text = "# " + taxon_title
        fig.add_trace(go.Bar(hovertemplate=hovertext, name="",x=x_values, y=list(n_OTUs.values()), showlegend=False))
        fig.update_traces(marker_color=color1, marker_line_color=color2, marker_line_width=1, opacity=opacity_value)
        fig.update_yaxes(title_text=title_text, title_standoff=5, rangemode="tozero")

        ## Number of OTUs on species level
        hovertext = 'Taxon: %{x}, Species: %{text}'
        fig.add_trace(go.Scatter(textposition = "top center", hovertemplate=hovertext, text=list(n_species.values()), name="Species",x=x_values, y=list(n_OTUs.values()), showlegend=False, mode='text'))
        fig.update_traces(marker_color=color1, marker_line_color=color2, marker_line_width=1, opacity=opacity_value)

        ## update the layout
        fig.update_layout(barmode='stack', uniformtext_mode='hide', height=int(height), width=int(width), template=template, showlegend=False, font_size=font_size, title_font_size=font_size)
        fig.update_xaxes(tickmode='linear')
        fig.update_xaxes(tickangle=-90)

        ## write ouput files
        output_pdf = Path(path_to_outdirs).joinpath('Per_taxon_statistics', '{}_{}_OTUs.pdf'.format(TaXon_table_xlsx.stem, taxonomic_level))
        output_html2 = Path(path_to_outdirs).joinpath('Per_taxon_statistics', '{}_{}_OTUs.html'.format(TaXon_table_xlsx.stem, taxonomic_level))
        fig.write_image(str(output_pdf))
        fig.write_html(str(output_html2))

        ## ask to show file
        answer = sg.PopupYesNo('Show plots?', keep_on_top=True)
        if answer == "Yes":
            webbrowser.open('file://' + str(output_html1))
            webbrowser.open('file://' + str(output_html2))

        ## print closing text
        closing_text = "Plots are found under: " + str(path_to_outdirs) + "/Per_taxon_statistics/"
        sg.Popup(closing_text, title="Finished", keep_on_top=True)

        ## write to log
        from taxontabletools.create_log import ttt_log
        ttt_log("per taxon statistics", "analysis", TaXon_table_xlsx.name, output_pdf.name, "", path_to_outdirs)



import random
import PySimpleGUI as sg
import pandas as pd
import numpy as np
import plotly.graph_objects as go
from pathlib import Path
import math, webbrowser
import plotly.express as px
from statistics import mean
from collections import Counter
from taxontabletools.taxontable_manipulation import strip_metadata
from taxontabletools.taxontable_manipulation import collect_metadata
from taxontabletools.taxontable_manipulation import add_metadata

def collect_reads(TaXon_table_xlsx):

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('')
    TaXon_table_df = strip_metadata(TaXon_table_df)
    samples = TaXon_table_df.columns.tolist()[10:]

    reads_per_sample = [sum(TaXon_table_df[sample].values.tolist()) for sample in samples]
    reads_sum = sum(reads_per_sample)
    reads_max = max(reads_per_sample)
    reads_mean = int(mean(reads_per_sample))
    reads_min = min(reads_per_sample)

    return [reads_sum, reads_max, reads_mean, reads_min]

def normalize_reads(TaXon_table_xlsx, path_to_outdirs, sub_sample_size):

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('')
    TaXon_table_df_metadata = collect_metadata(TaXon_table_df)
    TaXon_table_df = strip_metadata(TaXon_table_df)
    samples = TaXon_table_df.columns.tolist()[10:]

    ## count the lowest number of reads to use as threshold
    reads_list = []
    for sample in samples:
        reads_list.append(sum(TaXon_table_df[sample].values.tolist()))

    ## convert sub_sample_size to int
    sub_sample_size = int(sub_sample_size)

    ## collect all OTUs
    OTU_list = TaXon_table_df["ID"].values.tolist()

    df_out = pd.DataFrame()
    required_columns = TaXon_table_df.columns.tolist()[0:10]
    df_out[required_columns] = TaXon_table_df[required_columns]

    ############################################################################
    ## create the progress bar window
    layout = [[sg.Text('Progress bar')],
              [sg.ProgressBar(1000, orientation='h', size=(20, 20), key='progressbar')],
              [sg.Cancel()]]
    window_progress_bar = sg.Window('Progress bar', layout, keep_on_top=True)
    progress_bar = window_progress_bar['progressbar']
    progress_update = 0
    progress_increase = 1000 / len(samples)
    ############################################################################

    warnings = []

    for sample in samples:
        ## filter sample from data
        read_df = TaXon_table_df[[sample, "ID"]]
        ## drop empty OTUs
        read_df = read_df[read_df[sample] != 0]
        ## check if sample can be normalized, otherwise just keep all reads and OTUs
        if sub_sample_size <= sum(read_df[sample].values.tolist()):
            ## create read list to draw the subsamples from
            read_list = pd.Series(np.repeat(read_df['ID'].to_list(), read_df[sample].to_list()))
            ## create a random subsample
            sub_sample = read_list.sample(n = sub_sample_size)
            ## count the number of reads per OTU
            sub_sample_reads = dict(pd.Series(sub_sample).value_counts())
            ## create a sorted OTU list
            OTU_sample_list = []
            for OTU in OTU_list:
                if OTU in sub_sample_reads.keys():
                    OTU_sample_list.append(sub_sample_reads[OTU])
                else:
                    OTU_sample_list.append(0)
        else:
            OTU_sample_list = TaXon_table_df[[sample]]
            warnings.append(sample)

        ## add OTUs to dataframe
        df_out[sample] = OTU_sample_list

        ############################################################################
        event, values = window_progress_bar.read(timeout=1)
        if event == 'Cancel'  or event is None:
            window_progress_bar.Close()
            raise RuntimeError
        # update bar with loop value +1 so that bar eventually reaches the maximum
        progress_update += progress_increase
        progress_bar.UpdateBar(progress_update)
        ############################################################################

    window_progress_bar.Close()

    ## remove empty OTUs
    header = df_out.columns.tolist()
    row_filter_list = []
    for row in df_out.values.tolist():
        reads = sum(row[10:])
        if reads != 0:
            row_filter_list.append(row)
        else:
            print('Removed: {}'.format(row[0]))

    df_out = pd.DataFrame(row_filter_list)
    df_out.columns = header

    ## add already existing metadata back to the df
    if len(TaXon_table_df_metadata.columns) != 1:
        df_out = add_metadata(df_out, TaXon_table_df_metadata)

    ## write output file
    output_pdf = Path(str(path_to_outdirs) + "/" + "TaXon_tables" + "/" + TaXon_table_xlsx.stem + "_normalized.xlsx")
    df_out.to_excel(output_pdf, index=False)

    ## also count stats if selected
    OTU_loss = []
    for sample in samples:
        original_OTUs = len([i[1] for i in TaXon_table_df[[sample, "ID"]].values.tolist() if i[0] != 0])
        subsample_OTUs = len([i[1] for i in df_out[[sample, "ID"]].values.tolist() if i[0] != 0])

        if subsample_OTUs == 0:
            d_OTUs = 100
        else:
            d_OTUs = round(100 - subsample_OTUs / original_OTUs * 100, 4)
        OTU_loss.append(d_OTUs)

    closing_text = "All samples were normalized to maximum {} reads!\n\nMean discarded OTUs: {} %.\n\nSamples below the normalization theshold: {}".format(str(sub_sample_size), str(round(mean(OTU_loss), 4)), len(warnings))
    sg.Popup(closing_text, title='Finished')

    ## write to log
    from taxontabletools.create_log import ttt_log
    ttt_log("normalization", "processing", TaXon_table_xlsx.name, output_pdf.name, "", path_to_outdirs)









            #



import PySimpleGUI as sg
import pandas as pd
from pandas import DataFrame
import numpy as np
from pathlib import Path
import sys, subprocess, os

def convert_to_perlodes(TaXon_table_xlsx, operational_taxon_list, path_to_outdirs, meta_data_to_test):

    def open_table(table):
        if sys.platform == "win32":
            os.startfile(table)
        else:
            opener = "open" if sys.platform == 'darwin' else 'xdg-open'
            subprocess.call([opener, table])

    #get the taxonomy from the operational taxon list
    operational_taxon_list_df = pd.read_excel(Path(operational_taxon_list), header=2, sheet_name="Operationelle Taxaliste")
    taxonomy_list = operational_taxon_list_df["Taxonname\n(Perlodes-Datenbank)"].values.tolist()
    taxonomy_list = [x for x in taxonomy_list if str(x) != 'nan']

    # get the according IDs from the operational taxon list
    IDs_list = operational_taxon_list_df["ID_\nART"].values.tolist()
    IDs_list = [x for x in IDs_list if str(x) != 'nan']

    # create a dict to store both the ID and the taxonomy
    operational_taxon_list_dict = {}
    for ID, taxonomy in zip(IDs_list, taxonomy_list):
       operational_taxon_list_dict[taxonomy] = int(ID)

    # load the taxon table and create a list
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx)
    TaXon_table_taxonomy = TaXon_table_df.columns.tolist()[0:7]
    samples_list = TaXon_table_df.columns.tolist()[10:]

    ## load the metadata -> freshwater type
    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0).fillna("nan")
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()
    metadata_loc = Meta_data_table_df.columns.tolist().index(meta_data_to_test)
    types_list = Meta_data_table_df[meta_data_to_test].values.tolist()
    ## drop samples with metadata called nan (= empty)
    drop_samples = [i[0] for i in Meta_data_table_df.values.tolist() if i[metadata_loc] == "nan"]

    if drop_samples != []:
        sg.PopupError("Please fill out all the metadata for all samples.")

    elif sorted(Meta_data_table_samples) == sorted(samples_list):
        ## use the metadata sample sorting
        samples_list = sorted(Meta_data_table_samples)

        # store hits and dropped OTUs
        hit_list, dropped_list, transversion_list = [], [], []

        # loop through the taxon table
        for taxonomy in TaXon_table_df[TaXon_table_taxonomy].drop_duplicates().values.tolist():

           # collect the OTU name, species, genus and family and convert to perlodes format
           OTU = taxonomy[0]
           Species = taxonomy[6]
           Species_group = str(taxonomy[6]) + "-Gr."
           Genus = str(taxonomy[5]) + " sp."
           Family = str(taxonomy[4]) + " Gen. sp."

           # test if the OTU has a hit at: Species level, Genus level or Family level
           if Species in operational_taxon_list_dict.keys():
               # add to hit list
               hit_list.append([OTU] + [str(operational_taxon_list_dict[Species])] + [Species])
               # add to perlodes log file
               transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Species])] + [Species])

           elif Species_group in operational_taxon_list_dict.keys():
               hit_list.append([OTU] + [str(operational_taxon_list_dict[Species_group])] + [Species_group])
               transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Species_group])] + [Species_group])

           elif Genus in operational_taxon_list_dict.keys():
               hit_list.append([OTU] + [str(operational_taxon_list_dict[Genus])] + [Genus])
               transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Genus])] + [Genus])

           elif Family in operational_taxon_list_dict.keys():
               hit_list.append([OTU] + [str(operational_taxon_list_dict[Family])] + [Family])
               transversion_list.append(taxonomy + [str(operational_taxon_list_dict[Family])] + [Family])

           # otherwise store the hit with an "nan"
           else:
               hit_list.append([OTU] + ["nan"])
               dropped_list.append(OTU)
               transversion_list.append(taxonomy + ["", ""])

        # create an output list for perlodes
        # make read abundaces binary
        perlodes_input_list = []
        for hit, row in zip(hit_list, TaXon_table_df[samples_list].values.tolist()):
           # skip OTUs that were not in the OPT
           if hit[1] != "nan":
               reads_list = []
               # loop through all the sample of the file
               for reads in row:
                   # reads > 0 --> 1
                   if reads > 0:
                       reads_list.append(1)
                   # reads == 0 --> 0
                   else:
                       reads_list.append(0)
               # now append the taxonomy and the presence/absence to the perlodes list
               perlodes_input_list.append([hit[1]] + [hit[2]] + reads_list)

        # calculate the number of dropped OTUs
        number_of_initial_OTUs = len(TaXon_table_df[TaXon_table_taxonomy].drop_duplicates().values.tolist())

        # write the perlodes output file
        perlodes_df = pd.DataFrame(perlodes_input_list)
        perlodes_df.columns = ["ID_ART", "TAXON_NAME"] + samples_list

        # Perlocdes sums up the counts of all df_duplicates! So these must be removed
        perlodes_df.index = perlodes_df["TAXON_NAME"]
        # create a set of all present taxa
        taxa_set =  set(perlodes_df["TAXON_NAME"].values.tolist())
        perlodes_filtered_list = []

        # loop through all target taxa
        for taxon in taxa_set:
           # collect the ID and Taxon name
           ID_taxon_name = perlodes_df.loc[taxon][["ID_ART", "TAXON_NAME"]].values.tolist()[0]
           # if there are duplicates:
           try:
               # calculate the sum for each sample
               sum_of_pa = list(perlodes_df.loc[taxon][samples_list].sum())
               # start a new list for each taxon
               pa_list = ID_taxon_name
               # replace the count by either 1 or 0
               for pa in sum_of_pa:
                   if pa > 0:
                       pa_list.append(1)
                   else:
                       pa_list.append(0)
               perlodes_filtered_list.append(pa_list)
           # if there are no duplicates
           except:
               # simply take the old presence/absence line
               perlodes_filtered_list.append(perlodes_df.loc[taxon].values.tolist())

        ## create header rows for the input file
        a = [["Gewässertyp", ""] + types_list]
        b = [["Taxaliste", ""] + ["original"] * len(samples_list)]
        c = [["Nutzung", ""] + ["keine"] * len(samples_list)]

        # write the filtered list to a dataframe
        perlodes_df = pd.DataFrame(a + b + c + perlodes_filtered_list)
        perlodes_df.columns = ["ID_ART", "TAXON_NAME"] + samples_list
        perlodes_directory = Path(str(path_to_outdirs) + "/" + "Perlodes" + "/" + TaXon_table_xlsx.stem)
        perlodes_xlsx = Path(str(perlodes_directory) + "_perlodes.xlsx")
        perlodes_df.to_excel(perlodes_xlsx, sheet_name='ImportList', index=False)

        # write the log file file to a different dataframe
        transversion_df = pd.DataFrame(transversion_list)
        transversion_df.columns = ["ID", "Phylum", "Class", "Order", "Family", "Genus", "Species", "ID_ART", "TAXON_NAME"]
        transversion_xlsx = Path(str(perlodes_directory) + "_perlodes_conversion_table.xlsx")
        transversion_df.to_excel(transversion_xlsx, index=False)

        closing_text = "Perlodes input file is found under:\n" + '/'.join(str(perlodes_xlsx).split("/")[-4:])
        info_text = "\n\nDropped " + str(len(dropped_list)) + " of " + str(number_of_initial_OTUs) + " OTUs.\n"
        answer = sg.PopupYesNo(closing_text + info_text + "\nOpen table?")
        if answer == "Yes":
            open_table(perlodes_xlsx)

        from taxontabletools.create_log import ttt_log
        ttt_log("perlodes conversion", "processing", TaXon_table_xlsx.name, perlodes_xlsx.name, "nan", path_to_outdirs)

    else:
        sg.PopupError("Error: The samples from the metadata table do not match!")



import pandas as pd
from pathlib import Path
import PySimpleGUI as sg
import numpy as np
from taxontabletools.taxontable_manipulation import strip_metadata
from taxontabletools.taxontable_manipulation import collect_metadata
from taxontabletools.taxontable_manipulation import add_metadata

def rename_samples(TaXon_table_xlsx, meta_data_to_test, path_to_outdirs):

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('')
    TaXon_table_df_metadata = collect_metadata(TaXon_table_df)
    TaXon_table_df = strip_metadata(TaXon_table_df)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0).fillna("nan")
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()
    new_names = Meta_data_table_df[meta_data_to_test].values.tolist()

    if len(set(new_names)) != len(new_names):
        sg.PopupError("Error: Found duplicates in the new sample names!", keep_on_top=True)
    else:
        answer = sg.PopupOKCancel("Warning: This will overwrite your Taxon table and rename all your samples!\n\nProceed anyways?")

        if answer == "OK":
            if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):
                if len(set(Meta_data_table_samples)) == len(Meta_data_table_samples):
                    ## collect new names and rename the dataframe
                    columns_list = TaXon_table_df.columns.tolist()

                    ## relabel samples according to their index
                    for sample, rename in Meta_data_table_df[["Samples", meta_data_to_test]].values.tolist():
                        i = columns_list.index(sample)
                        columns_list[i] = rename

                    # save the renamed taxon table
                    TaXon_table_df.columns = columns_list
                    Meta_data_table_df["Old names"] = Meta_data_table_samples
                    Meta_data_table_df["Samples"] = new_names

                    ## add already existing metadata back to the df
                    if len(TaXon_table_df_metadata.columns) != 1:
                        TaXon_table_df = add_metadata(TaXon_table_df, TaXon_table_df_metadata)

                    ## write dataframes
                    TaXon_table_df.to_excel(TaXon_table_xlsx, index=False, sheet_name="TaXon table")
                    Meta_data_table_df.to_excel(Meta_data_table_xlsx, index=False)

                    sg.Popup("Samples were successfully renamed according to the metadata table!")
                else:
                    sg.PopupError("Error: Found duplicates in the sample names!", keep_on_top=True)
            else:
                sg.PopupError("Error: The samples between the taxon table and meta table do not match!", keep_on_top=True)

def sort_samples(TaXon_table_xlsx, path_to_outdirs):

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('')
    TaXon_table_df_metadata = collect_metadata(TaXon_table_df)
    TaXon_table_df = strip_metadata(TaXon_table_df)
    TaXon_table_samples = TaXon_table_df.columns.tolist()[10:]

    Meta_data_table_xlsx = Path(str(path_to_outdirs) + "/" + "Meta_data_table" + "/" + TaXon_table_xlsx.stem + "_metadata.xlsx")
    Meta_data_table_df = pd.read_excel(Meta_data_table_xlsx, header=0).fillna("nan")
    Meta_data_table_samples = Meta_data_table_df['Samples'].tolist()

    answer = sg.PopupOKCancel("Warning: This will overwrite your Taxon table and rename all your samples!\n\nProceed anyways?")

    if answer == "OK":
        if sorted(TaXon_table_samples) == sorted(Meta_data_table_samples):
            if len(set(TaXon_table_samples)) == len(TaXon_table_samples):
                new_df = pd.DataFrame()

                ## collect new sample names
                for sample in Meta_data_table_samples:
                    new_df[sample] = TaXon_table_df[sample].values.tolist()

                ## merge dataframes
                df1 = TaXon_table_df[TaXon_table_df.columns.tolist()[:10]]
                result = pd.concat([df1, new_df], axis=1, sort=False)

                ## add already existing metadata back to the df
                if len(TaXon_table_df_metadata.columns) != 1:
                    result = add_metadata(result, TaXon_table_df_metadata)

                ## write dataframe
                result.to_excel(TaXon_table_xlsx, index=False)
                sg.Popup("Taxon table was successfully sorted according to the metadata table!")
            else:
                sg.PopupError("Error: Found duplicates in the sample names!", keep_on_top=True)
        else:
            sg.PopupError("Error: The samples between the taxon table and meta table do not match!", keep_on_top=True)












#



import pandas as pd
import plotly.graph_objects as go
from taxontabletools.taxontable_manipulation import strip_metadata
from taxontabletools.taxontable_manipulation import collect_metadata
from taxontabletools.taxontable_manipulation import add_metadata
from taxontabletools.taxontable_manipulation import reduce_taxontable
import PySimpleGUI as sg
from pathlib import Path

def itt_import(TaXon_table_xlsx, ITT_table_xlsx, path_to_outdirs):

    ## read itt sheet
    itt_df = pd.read_excel(Path(ITT_table_xlsx))
    ## rename columns
    itt_df.columns = itt_df.values.tolist()[0]
    ## select trait data
    itt_df = itt_df[1:587]

    ## collect all present families
    itt_families = itt_df['family'].values.tolist()

    ## create a list of all available trait categories
    trait_categories = itt_df.columns.tolist()[3:]

    ## create a dictionary to store the trait of each family
    family_traits = {}

    for family in itt_df.values.tolist():
        ## collect family name
        family_name = family[1]
        ## store traits
        family_traits[family_name] = family[3:]

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('')
    TaXon_table_df_metadata = collect_metadata(TaXon_table_df)
    TaXon_table_df = strip_metadata(TaXon_table_df)

    OTU_list = []

    for OTU in TaXon_table_df.values.tolist():
        ## collect family name
        family_name = OTU[4]

        traits_list = []

        larva_habitat_dict = {}

        if family_name in family_traits.keys():

            ## TRAIT 1
            larva_habitat = ['terrestrial', 'aquatic']
            t1 = family_traits[family_name][0:2]
            lst = []
            for i,j in enumerate(t1):
                if t1[i] > 0:
                    lst.append(larva_habitat[i])
            traits_list.append('+'.join(lst))

            ## TRAIT 2
            adult_habitat = ['terrestrial', 'aquatic']
            t1 = family_traits[family_name][2:4]
            lst = []
            for i,j in enumerate(t1):
                if t1[i] > 0:
                    lst.append(adult_habitat[i])
            traits_list.append('+'.join(lst))

            ## TRAIT 3
            larva_diet_types = ['phytophagous', 'zoophagous', 'mycetophagous', 'saprophagous', 'detritophagous', 'coprophagous']
            t1 = family_traits[family_name][4:10]
            lst = []
            for i,j in enumerate(t1):
                if t1[i] > 0:
                    lst.append(larva_diet_types[i])
            traits_list.append('+'.join(lst))

            ## TRAIT 4
            larva_diet_types_phytophageous = ['phyllophagous', 'saproxylic', 'sap-sucking', 'stem', 'flower', 'seed', 'gall-inducing', 'miners', 'roots']
            t1 = family_traits[family_name][10:19]
            lst = []
            for i,j in enumerate(t1):
                if t1[i] > 0:
                    lst.append(larva_diet_types_phytophageous[i])
            traits_list.append('+'.join(lst))

            ## TRAIT 5
            larva_diet_types_zoophagus = ['predator', 'micropredator', 'parasite', 'parasitoid', 'necrophorous']
            t1 = family_traits[family_name][20:25]
            lst = []
            for i,j in enumerate(t1):
                if t1[i] > 0:
                    lst.append(larva_diet_types_zoophagus[i])
            traits_list.append('+'.join(lst))

            OTU_list.append(OTU[0:9] + traits_list + OTU[9:])

        else:
            OTU_list.append(OTU[0:9] + ['']*5 + OTU[9:])

    ## create a new dataframe including the ITT traits
    new_colums = TaXon_table_df.columns.tolist()[0:9] + ['larva_habitat', 'adult_habitat', 'larva_diet_types', 'larva_diet_types_phytophageous', 'larva_diet_types_zoophagus'] + TaXon_table_df.columns.tolist()[9:]

    new_df = pd.DataFrame(OTU_list, columns=new_colums)

    ## add already existing metadata back to the df
    if len(TaXon_table_df_metadata.columns) != 1:
        new_df = add_metadata(new_df, TaXon_table_df_metadata)

    ## save dataframe
    output_pdf = Path(str(path_to_outdirs) + "/TaXon_tables/" + TaXon_table_xlsx.stem + "_itt.xlsx")
    new_df.to_excel(output_pdf, index=False)

    sg.Popup('All ITT traits have been added to the Taxon table.')


def itt_stats(path_to_outdirs, metadata_to_test, taxonomic_level):

    ## define level to test
    taxonomic_level = 'OTUs'
    meta_data_to_test = 'larva_diet_types_zoophagus'

    ## load TaxonTable
    TaXon_table_xlsx = Path(TaXon_table_xlsx)
    TaXon_table_df = pd.read_excel(TaXon_table_xlsx).fillna('nan')
    ## collect samples and metadata
    samples = strip_metadata(TaXon_table_df).columns.tolist()[10:]
    categories = set(TaXon_table_df[meta_data_to_test].values.tolist())

    ## adjust taxonomic level if neccessary
    taxon_title = taxonomic_level
    if taxonomic_level in ["ASVs", "ESVs", "OTUs", "zOTUs"]:
        taxon_title = taxonomic_level
        taxonomic_level = "ID"

    ## reduce df to samples, taxon and metadata
    TaXon_table_df = reduce_taxontable(TaXon_table_df, taxonomic_level, samples, meta_data_to_test)

    ## loop through categories and collect number of occurrences per category
    categories_dict = {}
    for category in categories:
        y_sample_counts = []
        for sample in samples:
            present_categories = [i[2] for i in TaXon_table_df[[taxonomic_level, sample, meta_data_to_test]].values.tolist() if i[1] != 0]
            y_sample_counts.append(present_categories.count(category))
        categories_dict[category] = y_sample_counts

    ## loop through samples and collect the number of taxa
    n_taxa_total = []
    for sample in samples:
        n_taxa_total.append(len([i for i in TaXon_table_df[[taxonomic_level, sample]].values.tolist() if i[1] != 0]))

    ## initiate a new figure
    fig = go.Figure()

    ## calculate relative abundance per category
    for key,values in categories_dict.items():
        rel_values = [j/n_taxa_total[i]*100 for i,j in enumerate(values)]

        if key == 'nan':
            key = 'not specified'
        fig = fig.add_trace(go.Bar(x=samples, y=rel_values, name=str(key)))


    fig.update_layout(barmode='stack', width=2000, height=800, template='plotly_white', title=str(meta_data_to_test))
    fig.update_xaxes(tickmode='linear')
    fig.update_yaxes(title=taxon_title + ' (%)')
    fig.show()


#
